from __future__ import annotations

import logging
import typing as t
import uuid

from globus_sdk._missing import MISSING, MissingType
from globus_sdk.authorizers import NullAuthorizer
from globus_sdk.response import GlobusHTTPResponse
from globus_sdk.scopes import Scope
from globus_sdk.transport import RequestsTransport, RetryConfig

from ..flow_managers import GlobusNativeAppFlowManager
from ..response import OAuthRefreshTokenResponse
from .base_login_client import AuthLoginClient

log = logging.getLogger(__name__)


class NativeAppAuthClient(AuthLoginClient):
    """
    This type of ``AuthLoginClient`` is used to represent a Native App's
    communications with Globus Auth.
    It requires a Client ID, and cannot take an ``authorizer``.

    Native Apps are applications, like the Globus CLI, which are run
    client-side and therefore cannot keep secrets. Unable to possess client
    credentials, several Globus Auth interactions have to be specialized to
    accommodate the absence of a secret.

    Any keyword arguments given are passed through to the ``AuthClient``
    constructor.

    .. automethodlist:: globus_sdk.NativeAppAuthClient
    """

    def __init__(
        self,
        client_id: uuid.UUID | str,
        environment: str | None = None,
        base_url: str | None = None,
        app_name: str | None = None,
        transport: RequestsTransport | None = None,
        retry_config: RetryConfig | None = None,
    ) -> None:
        super().__init__(
            client_id=client_id,
            authorizer=NullAuthorizer(),
            environment=environment,
            base_url=base_url,
            app_name=app_name,
            transport=transport,
            retry_config=retry_config,
        )

    def oauth2_start_flow(
        self,
        requested_scopes: str | Scope | t.Iterable[str | Scope],
        *,
        redirect_uri: str | MissingType = MISSING,
        state: str = "_default",
        verifier: str | MissingType = MISSING,
        refresh_tokens: bool = False,
        prefill_named_grant: str | MissingType = MISSING,
    ) -> GlobusNativeAppFlowManager:
        """
        Starts a Native App OAuth2 flow.

        This is done internally by instantiating a
        :class:`GlobusNativeAppFlowManager <.GlobusNativeAppFlowManager>`

        While the flow is in progress, the ``NativeAppAuthClient`` becomes
        non thread-safe as temporary state is stored during the flow.

        :param requested_scopes: The scopes on the token(s) being requested. Defaults to
            ``openid profile email urn:globus:auth:scope:transfer.api.globus.org:all``
        :param redirect_uri: The page that users should be directed to after
            authenticating at the authorize URL. Defaults to
            'https://auth.globus.org/v2/web/auth-code', which displays the resulting
            ``auth_code`` for users to copy-paste back into your application (and
            thereby be passed back to the ``GlobusNativeAppFlowManager``)
        :param state: The ``redirect_uri`` page will have this included in a query
            parameter, so you can use it to pass information to that page if you use a
            custom page. It defaults to the string '_default'
        :param verifier: A secret used for the Native App flow. It will by default be a
            freshly generated random string, known only to this
            ``GlobusNativeAppFlowManager`` instance
        :param refresh_tokens: When True, request refresh tokens in addition to access
            tokens. [Default: ``False``]
        :param prefill_named_grant: Prefill the named grant label on the consent page

        .. tab-set::

            .. tab-item:: Example Usage

                You can see an example of this flow :ref:`in the usage examples
                <examples_native_app_login>`.

            .. tab-item:: API Info

                The Globus Auth specification for Native App grants details
                modifications to the Authorization Code grant flow as
                `The PKCE Security Protocol
                <https://docs.globus.org/api/auth/developer-guide/#pkce>`_.
        """
        log.debug("Starting Native App Grant Flow")
        self.current_oauth2_flow_manager = GlobusNativeAppFlowManager(
            self,
            requested_scopes=requested_scopes,
            redirect_uri=redirect_uri,
            state=state,
            verifier=verifier,
            refresh_tokens=refresh_tokens,
            prefill_named_grant=prefill_named_grant,
        )
        return self.current_oauth2_flow_manager

    def oauth2_refresh_token(
        self,
        refresh_token: str,
        *,
        body_params: dict[str, t.Any] | None = None,
    ) -> OAuthRefreshTokenResponse:
        """
        ``NativeAppAuthClient`` specializes the refresh token grant to include
        its client ID as a parameter in the POST body.
        It needs this specialization because it cannot authenticate the refresh
        grant call with client credentials, as is normal.

        :param refresh_token: The refresh token to use to get a new access token
        :param body_params: Extra parameters to include in the POST body
        """
        log.debug("Executing token refresh without client credentials")
        form_data = {
            "refresh_token": refresh_token,
            "grant_type": "refresh_token",
            "client_id": self.client_id,
        }
        return self.oauth2_token(
            form_data, body_params=body_params, response_class=OAuthRefreshTokenResponse
        )

    def create_native_app_instance(
        self,
        template_id: uuid.UUID | str,
        name: str,
    ) -> GlobusHTTPResponse:
        """
         Create a new native app instance. The new instance is a confidential client.

         :param template_id: The client ID of the calling native app
         :param name: The name given to the new app instance

        .. tab-set::

             .. tab-item:: Example Usage

                 .. code-block:: pycon

                     >>> ac = globus_sdk.NativeAppAuthClient(...)
                     >>> template_id = ...
                     >>> r = ac.create_native_app_instance(
                     ...     template_id,
                     ...     "My new native app instance",
                     ... )
                     >>> client_id = r["client"]["id"]

             .. tab-item:: Example Response Data

                 .. expandtestfixture:: auth.create_native_app_instance

             .. tab-item:: API Info

                 ``POST /v2/api/clients``

                 .. extdoclink:: Create Client
                     :ref: auth/reference/#create_client
        """
        body: dict[str, t.Any] = {
            "name": name,
            "template_id": template_id,
        }
        return self.post("/v2/api/clients", data={"client": body})



from __future__ import annotations

import textwrap
import typing as t
from contextlib import contextmanager

import globus_sdk
from globus_sdk._internal.utils import get_nice_hostname
from globus_sdk.exc.base import GlobusError
from globus_sdk.gare import GlobusAuthorizationParameters

from .login_flow_manager import LoginFlowManager

if t.TYPE_CHECKING:
    from globus_sdk.globus_app import GlobusAppConfig


class CommandLineLoginFlowEOFError(GlobusError, EOFError):
    """
    An error raised when a CommandLineLoginFlowManager reads an EOF when prompting for
    a code.
    """


class CommandLineLoginFlowManager(LoginFlowManager):
    """
    A login flow manager which drives authorization-code token grants through the
    command line.

    :param AuthLoginClient login_client: The client that will be making Globus
        Auth API calls required for a login flow.

        .. note::
            If this client is a :class:`globus_sdk.ConfidentialAppAuthClient`, an
            explicit `redirect_uri` param is required.

    :param str redirect_uri: The redirect URI to use for the login flow. When the
        `login_client` is a native client, this defaults to a Globus-hosted URL.
    :param bool request_refresh_tokens: A signal of whether refresh tokens are expected
        to be requested, in addition to access tokens.
    :param str native_prefill_named_grant: A string to prefill in a Native App login
        flow. This value is only used if the `login_client` is a native client.
    """

    def __init__(
        self,
        login_client: globus_sdk.AuthLoginClient,
        *,
        redirect_uri: str | None = None,
        request_refresh_tokens: bool = False,
        native_prefill_named_grant: str | None = None,
    ) -> None:
        super().__init__(
            login_client,
            request_refresh_tokens=request_refresh_tokens,
            native_prefill_named_grant=native_prefill_named_grant,
        )

        if redirect_uri is None:
            # Confidential clients must always define their own custom redirect URI.
            if isinstance(login_client, globus_sdk.ConfidentialAppAuthClient):
                msg = "Use of a Confidential client requires an explicit redirect_uri."
                raise globus_sdk.GlobusSDKUsageError(msg)

            # Native clients may infer the globus-provided helper page if omitted.
            redirect_uri = login_client.base_url + "v2/web/auth-code"
        self.redirect_uri = redirect_uri

    @classmethod
    def for_globus_app(
        cls,
        app_name: str,
        login_client: globus_sdk.AuthLoginClient,
        config: GlobusAppConfig,
    ) -> CommandLineLoginFlowManager:
        """
        Create a ``CommandLineLoginFlowManager`` for use in a GlobusApp.

        :param app_name: The name of the app. Will be prefilled in native auth flows.
        :param login_client: A client used to make Globus Auth API calls.
        :param config: A GlobusApp-bounded object used to configure login flow manager.
        :raises GlobusSDKUsageError: if login_redirect_uri is not set on the config
            but a ConfidentialAppAuthClient is supplied.
        """
        hostname = get_nice_hostname()
        if hostname:
            prefill = f"{app_name} on {hostname}"
        else:
            prefill = app_name

        return cls(
            login_client,
            redirect_uri=config.login_redirect_uri,
            request_refresh_tokens=config.request_refresh_tokens,
            native_prefill_named_grant=prefill,
        )

    def run_login_flow(
        self,
        auth_parameters: GlobusAuthorizationParameters,
    ) -> globus_sdk.OAuthTokenResponse:
        """
        Run an interactive login flow on the command line to get tokens for the user.

        :param auth_parameters: ``GlobusAuthorizationParameters`` passed through
            to the authentication flow to control how the user will authenticate.
        """
        authorize_url = self._get_authorize_url(auth_parameters, self.redirect_uri)
        self.print_authorize_url(authorize_url)
        auth_code = self.prompt_for_code()

        return self.login_client.oauth2_exchange_code_for_tokens(auth_code)

    def print_authorize_url(self, authorize_url: str) -> None:
        """
        Prompt the user to authenticate using the provided ``authorize_url``.

        :param authorize_url: The URL at which the user will login and consent to
            application accesses.
        """
        login_prompt = "Please authenticate with Globus here:"
        print(textwrap.dedent(f"""
                {login_prompt}
                {"-" * len(login_prompt)}
                {authorize_url}
                {"-" * len(login_prompt)}
                """))

    def prompt_for_code(self) -> str:
        """
        Prompt the user to enter an authorization code.

        :returns: The authorization code entered by the user.
        """
        code_prompt = "Enter the resulting Authorization Code here: "
        with self._handle_input_errors():
            return input(code_prompt).strip()

    @contextmanager
    def _handle_input_errors(self) -> t.Iterator[None]:
        try:
            yield
        except EOFError as e:
            msg = (
                "An EOF was read when an authorization code was expected."
                " (Are you running this in an interactive terminal?)"
            )
            raise CommandLineLoginFlowEOFError(msg) from e



#!/usr/bin/env python3

import datetime

import globus_sdk

intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
}

# pull signature docs from type hints, into body
# and keep the signatures concise
autodoc_typehints = "description"
# do not generate doc stubs for inherited parameters from a superclass,
# merely because they are type annotated
autodoc_typehints_description_target = "documented_params"


# sphinx extensions (minimally, we want autodoc and viewcode to build the site)
# plus, we have our own custom extension in the SDK to include
extensions = [
    # sphinx-included extensions
    "sphinx.ext.autodoc",
    "sphinx.ext.intersphinx",
    "sphinx.ext.viewcode",
    # other packages
    "sphinx_copybutton",
    "sphinx_design",
    "sphinx_issues",
    # our custom one
    "globus_sdk._internal.extensions.sphinxext",
]

project = "globus-sdk"
copyright = f"2016-{datetime.datetime.today().strftime('%Y')}, Globus"
author = "Globus Team"
# The short X.Y version.
version = globus_sdk.__version__
# The full version, including alpha/beta/rc tags.
release = version
major_version = version.partition(".")[0]

issues_github_path = "globus/globus-sdk-python"

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = "en"

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = ["_build"]


# HTML Theme Options
html_show_sourcelink = True
html_theme = "furo"
html_title = f"globus-sdk v{major_version}"
html_theme_options = {
    "light_css_variables": {
        "color-brand-primary": "#27518F",
    },
}
html_logo = "_static/logo.png"
html_static_path = ["_static"]
html_css_files = ["css/globus_sdk_tab_borders.css"]

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "friendly"
pygments_dark_style = "monokai"  # this is a furo-specific option

# Output file base name for HTML help builder.
htmlhelp_basename = "globus-sdk-doc"



from .add_content_directive import AddContentDirective
from .automethodlist import AutoMethodList
from .copy_params import CopyParams
from .enumerate_testing_fixtures import EnumerateTestingFixtures
from .expand_testing_fixture import ExpandTestingFixture
from .externaldoclink import ExternalDocLink
from .list_known_scopes import ListKnownScopes
from .paginated_usage import PaginatedUsage

__all__ = (
    "AddContentDirective",
    "AutoMethodList",
    "CopyParams",
    "EnumerateTestingFixtures",
    "ExpandTestingFixture",
    "ExternalDocLink",
    "ListKnownScopes",
    "PaginatedUsage",
)



from globus_sdk.testing import load_response


def test_get_policies(service_client):
    meta = load_response(service_client.get_policies).metadata
    res = service_client.get_policies()

    assert {policy["id"] for policy in res["policies"]} == set(meta["policy_ids"])



from __future__ import annotations

import logging
import typing as t
import uuid

from globus_sdk import GlobusHTTPResponse, client
from globus_sdk._internal.remarshal import strseq_listify
from globus_sdk._missing import MISSING, MissingType
from globus_sdk.scopes import ComputeScopes

from .errors import ComputeAPIError

log = logging.getLogger(__name__)


class ComputeClientV2(client.BaseClient):
    r"""
    Client for the Globus Compute API, version 2.

    .. sdk-sphinx-copy-params:: BaseClient

    .. automethodlist:: globus_sdk.ComputeClientV2
    """

    error_class = ComputeAPIError
    service_name = "compute"
    scopes = ComputeScopes
    default_scope_requirements = [ComputeScopes.all]

    def get_version(self, service: str | MissingType = MISSING) -> GlobusHTTPResponse:
        """Get the current version of the API and other services.

        :param service: Service for which to get version information.

        .. tab-set::

            .. tab-item:: API Info

                .. extdoclink:: Get Version
                    :service: compute
                    :ref: Root/operation/get_version_v2_version_get
        """
        query_params = {"service": service}
        return self.get("/v2/version", query_params=query_params)

    def get_result_amqp_url(self) -> GlobusHTTPResponse:
        """Generate new credentials (in the form of a connection URL) for
        connecting to the AMQP service.

        .. tab-set::

            .. tab-item:: API Info

                .. extdoclink:: Get Result AMQP URL
                    :service: compute
                    :ref: Root/operation/get_user_specific_result_amqp_url_v2_get_amqp_result_connection_url_get
        """  # noqa: E501
        return self.get("/v2/get_amqp_result_connection_url")

    def register_endpoint(self, data: dict[str, t.Any]) -> GlobusHTTPResponse:
        """Register a new endpoint.

        :param data: An endpoint registration document.

        .. tab-set::

            .. tab-item:: API Info

                .. extdoclink:: Register Endpoint
                    :service: compute
                    :ref: Endpoints/operation/register_endpoint_v2_endpoints_post
        """
        return self.post("/v2/endpoints", data=data)

    def get_endpoint(self, endpoint_id: uuid.UUID | str) -> GlobusHTTPResponse:
        """Get information about a registered endpoint.

        :param endpoint_id: The ID of the Globus Compute endpoint.

        .. tab-set::

            .. tab-item:: API Info

                .. extdoclink:: Get Endpoint
                    :service: compute
                    :ref: Endpoints/operation/get_endpoint_v2_endpoints__endpoint_uuid__get
        """  # noqa: E501
        return self.get(f"/v2/endpoints/{endpoint_id}")

    def get_endpoint_status(self, endpoint_id: uuid.UUID | str) -> GlobusHTTPResponse:
        """Get the status of a registered endpoint.

        :param endpoint_id: The ID of the Globus Compute endpoint.

        .. tab-set::

            .. tab-item:: API Info

                .. extdoclink:: Get Endpoint Status
                    :service: compute
                    :ref: Endpoints/operation/get_endpoint_status_v2_endpoints__endpoint_uuid__status_get
        """  # noqa: E501
        return self.get(f"/v2/endpoints/{endpoint_id}/status")

    def get_endpoints(self, role: str | MissingType = MISSING) -> GlobusHTTPResponse:
        """Get a list of registered endpoints associated with the authenticated user.

        :param role: Role of the user in relation to endpoints. (e.g.: owner, any)

        .. tab-set::

            .. tab-item:: API Info

                .. extdoclink:: Get Endpoints
                    :service: compute
                    :ref: Endpoints/operation/get_endpoints_v2_endpoints_get
        """  # noqa: E501
        query_params = {"role": role}
        return self.get("/v2/endpoints", query_params=query_params)

    def delete_endpoint(self, endpoint_id: uuid.UUID | str) -> GlobusHTTPResponse:
        """Delete a registered endpoint.

        :param endpoint_id: The ID of the Globus Compute endpoint.

        .. tab-set::

            .. tab-item:: API Info

                .. extdoclink:: Delete Endpoint
                    :service: compute
                    :ref: Endpoints/operation/delete_endpoint_v2_endpoints__endpoint_uuid__delete
        """  # noqa: E501
        return self.delete(f"/v2/endpoints/{endpoint_id}")

    def lock_endpoint(self, endpoint_id: uuid.UUID | str) -> GlobusHTTPResponse:
        """Temporarily block registration requests for the endpoint.

        :param endpoint_id: The ID of the Globus Compute endpoint.

        .. tab-set::

            .. tab-item:: API Info

                .. extdoclink:: Lock Endpoint
                    :service: compute
                    :ref: Endpoints/operation/lock_endpoint_v2_endpoints__endpoint_uuid__lock_post
        """  # noqa: E501
        return self.post(f"/v2/endpoints/{endpoint_id}/lock")

    def register_function(self, data: dict[str, t.Any]) -> GlobusHTTPResponse:
        """Register a new function.

        :param data: A function registration document.

        .. tab-set::

            .. tab-item:: API Info

                .. extdoclink:: Register Function
                    :service: compute
                    :ref: Functions/operation/register_function_v2_functions_post
        """  # noqa: E501
        return self.post("/v2/functions", data=data)

    def get_function(self, function_id: uuid.UUID | str) -> GlobusHTTPResponse:
        """Get information about a registered function.

        :param function_id: The ID of the function.

        .. tab-set::

            .. tab-item:: API Info

                .. extdoclink:: Get Function
                    :service: compute
                    :ref: Functions/operation/get_function_v2_functions__function_uuid__get
        """  # noqa: E501
        return self.get(f"/v2/functions/{function_id}")

    def delete_function(self, function_id: uuid.UUID | str) -> GlobusHTTPResponse:
        """Delete a registered function.

        :param function_id: The ID of the function.

        .. tab-set::

            .. tab-item:: API Info

                .. extdoclink:: Delete Function
                    :service: compute
                    :ref: Functions/operation/delete_function_v2_functions__function_uuid__delete
        """  # noqa: E501
        return self.delete(f"/v2/functions/{function_id}")

    def get_task(self, task_id: uuid.UUID | str) -> GlobusHTTPResponse:
        """Get information about a task.

        :param task_id: The ID of the task.

        .. tab-set::

            .. tab-item:: API Info

                .. extdoclink:: Get Task
                    :service: compute
                    :ref: Tasks/operation/get_task_status_and_result_v2_tasks__task_uuid__get
        """  # noqa: E501
        return self.get(f"/v2/tasks/{task_id}")

    def get_task_batch(
        self, task_ids: uuid.UUID | str | t.Iterable[uuid.UUID | str]
    ) -> GlobusHTTPResponse:
        """Get information about a batch of tasks.

        :param task_ids: The IDs of the tasks.

        .. tab-set::

            .. tab-item:: API Info

                .. extdoclink:: Get Task Batch
                    :service: compute
                    :ref: Root/operation/get_batch_status_v2_batch_status_post
        """
        return self.post(
            "/v2/batch_status", data={"task_ids": strseq_listify(task_ids)}
        )

    def get_task_group(self, task_group_id: uuid.UUID | str) -> GlobusHTTPResponse:
        """Get a list of task IDs associated with a task group.

        :param task_group_id: The ID of the task group.

        .. tab-set::

            .. tab-item:: API Info

                .. extdoclink:: Get Task Group Tasks
                    :service: compute
                    :ref: TaskGroup/operation/get_task_group_tasks_v2_taskgroup__task_group_uuid__get
        """  # noqa: E501
        return self.get(f"/v2/taskgroup/{task_group_id}")

    def submit(self, data: dict[str, t.Any]) -> GlobusHTTPResponse:
        """Submit a batch of tasks to a Globus Compute endpoint.

        :param data: The task batch document.

        .. tab-set::

            .. tab-item:: API Info

                .. extdoclink:: Submit Batch
                    :service: compute
                    :ref: Root/operation/submit_batch_v2_submit_post
        """  # noqa: E501
        return self.post("/v2/submit", data=data)


class ComputeClientV3(client.BaseClient):
    r"""
    Client for the Globus Compute API, version 3.

    .. automethodlist:: globus_sdk.ComputeClientV3
    """

    error_class = ComputeAPIError
    service_name = "compute"
    scopes = ComputeScopes
    default_scope_requirements = [ComputeScopes.all]

    def register_endpoint(self, data: dict[str, t.Any]) -> GlobusHTTPResponse:
        """Register a new endpoint.

        :param data: An endpoint registration document.

        .. tab-set::

            .. tab-item:: API Info

                .. extdoclink:: Register Endpoint
                    :service: compute
                    :ref: Endpoints/operation/register_endpoint_v3_endpoints_post
        """
        return self.post("/v3/endpoints", data=data)

    def update_endpoint(
        self, endpoint_id: uuid.UUID | str, data: dict[str, t.Any]
    ) -> GlobusHTTPResponse:
        """Update an endpoint.

        :param endpoint_id: The ID of the endpoint.
        :param data: An endpoint update document.

        .. tab-set::

            .. tab-item:: API Info

                .. extdoclink:: Update Endpoint
                    :service: compute
                    :ref: Endpoints/operation/update_endpoint_v3_endpoints__endpoint_uuid__put
        """  # noqa: E501
        return self.put(f"/v3/endpoints/{endpoint_id}", data=data)

    def lock_endpoint(self, endpoint_id: uuid.UUID | str) -> GlobusHTTPResponse:
        """Temporarily block registration requests for the endpoint.

        :param endpoint_id: The ID of the Globus Compute endpoint.

        .. tab-set::

            .. tab-item:: API Info

                .. extdoclink:: Lock Endpoint
                    :service: compute
                    :ref: Endpoints/operation/lock_endpoint_v3_endpoints__endpoint_uuid__lock_post
        """  # noqa: E501
        return self.post(f"/v3/endpoints/{endpoint_id}/lock")

    def get_endpoint_allowlist(
        self, endpoint_id: uuid.UUID | str
    ) -> GlobusHTTPResponse:
        """Get a list of IDs for functions allowed to run on an endpoint.

        :param endpoint_id: The ID of the Globus Compute endpoint.

        .. tab-set::

            .. tab-item:: API Info

                .. extdoclink:: Get Endpoint Allowlist
                    :service: compute
                    :ref: Endpoints/operation/get_endpoint_allowlist_v3_endpoints__endpoint_uuid__allowed_functions_get
        """  # noqa: E501
        return self.get(f"/v3/endpoints/{endpoint_id}/allowed_functions")

    def register_function(self, data: dict[str, t.Any]) -> GlobusHTTPResponse:
        """Register a new function.

        :param data: A function registration document.

        .. tab-set::

            .. tab-item:: API Info

                .. extdoclink:: Register Function
                    :service: compute
                    :ref: Functions/operation/register_function_v3_functions_post
        """
        return self.post("/v3/functions", data=data)

    def submit(
        self, endpoint_id: uuid.UUID | str, data: dict[str, t.Any]
    ) -> GlobusHTTPResponse:
        """Submit a batch of tasks to a Globus Compute endpoint.

        :param endpoint_id: The ID of the Globus Compute endpoint.
        :param data: The task batch document.

        .. tab-set::

            .. tab-item:: API Info

                .. extdoclink:: Submit Batch
                    :service: compute
                    :ref: Endpoints/operation/submit_batch_v3_endpoints__endpoint_uuid__submit_post
        """  # noqa: E501
        return self.post(f"/v3/endpoints/{endpoint_id}/submit", data=data)



from .helpers import construct_error, get_last_request
from .models import RegisteredResponse, ResponseList, ResponseSet
from .registry import (
    get_response_set,
    load_response,
    load_response_set,
    register_response_set,
)

__all__ = (
    "get_last_request",
    "construct_error",
    "ResponseSet",
    "ResponseList",
    "RegisteredResponse",
    "load_response_set",
    "load_response",
    "get_response_set",
    "register_response_set",
)



from __future__ import annotations

import logging
import typing as t
import uuid

from cryptography.hazmat.primitives.asymmetric.rsa import RSAPublicKey

from globus_sdk import client, exc
from globus_sdk._internal.remarshal import commajoin, strseq_listify
from globus_sdk._missing import MISSING, MissingType
from globus_sdk.authorizers import GlobusAuthorizer
from globus_sdk.response import GlobusHTTPResponse, IterableResponse
from globus_sdk.scopes import AuthScopes, Scope
from globus_sdk.transport import RequestsTransport, RetryConfig

if t.TYPE_CHECKING:
    from globus_sdk.globus_app import GlobusApp

from .._common import get_jwk_data, pem_decode_jwk_data
from ..data import DependentScopeSpec
from ..errors import AuthAPIError
from ..response import (
    GetClientCredentialsResponse,
    GetClientsResponse,
    GetConsentsResponse,
    GetIdentitiesResponse,
    GetIdentityProvidersResponse,
    GetPoliciesResponse,
    GetProjectsResponse,
    GetScopesResponse,
)

log = logging.getLogger(__name__)

F = t.TypeVar("F", bound=t.Callable[..., GlobusHTTPResponse])


class AuthClient(client.BaseClient):
    """
    A client for using the
    `Globus Auth API <https://docs.globus.org/api/auth/>`_

    .. sdk-sphinx-copy-params:: BaseClient

    This class provides helper methods for most common resources in the
    Auth API, and the common low-level interface from
    :class:`BaseClient <globus_sdk.client.BaseClient>` of ``get``, ``put``,
    ``post``, and ``delete`` methods, which can be used to access any API
    resource.

    **Examples**

    Initializing an ``AuthClient`` to authenticate a user making calls to the
    Globus Auth service with an access token takes the form

    >>> from globus_sdk import AuthClient, AccessTokenAuthorizer
    >>> ac = AuthClient(authorizer=AccessTokenAuthorizer('<token_string>'))

    Other authorizers, most notably ``RefreshTokenAuthorizer``, are also supported.

    .. automethodlist:: globus_sdk.AuthClient
    """

    service_name = "auth"
    error_class = AuthAPIError
    scopes = AuthScopes
    default_scope_requirements = [
        AuthScopes.openid,
        AuthScopes.profile,
        AuthScopes.email,
    ]

    def __init__(
        self,
        environment: str | None = None,
        base_url: str | None = None,
        app: GlobusApp | None = None,
        app_scopes: list[Scope] | None = None,
        authorizer: GlobusAuthorizer | None = None,
        app_name: str | None = None,
        transport: RequestsTransport | None = None,
        retry_config: RetryConfig | None = None,
    ) -> None:
        super().__init__(
            environment=environment,
            base_url=base_url,
            app=app,
            app_scopes=app_scopes,
            authorizer=authorizer,
            app_name=app_name,
            transport=transport,
            retry_config=retry_config,
        )

    # FYI: this get_openid_configuration method is duplicated in AuthLoginBaseClient
    # if this code is modified, please update that copy as well
    # this will ideally be resolved in a future SDK version by making this the only copy
    def get_openid_configuration(self) -> GlobusHTTPResponse:
        """
        Fetch the OpenID Connect configuration data from the well-known URI for Globus
        Auth.
        """
        log.debug("Fetching OIDC Config")
        return self.get("/.well-known/openid-configuration")

    @t.overload
    def get_jwk(
        self,
        openid_configuration: GlobusHTTPResponse | dict[str, t.Any] | MissingType,
        *,
        as_pem: t.Literal[True],
    ) -> RSAPublicKey: ...

    @t.overload
    def get_jwk(
        self,
        openid_configuration: GlobusHTTPResponse | dict[str, t.Any] | MissingType,
        *,
        as_pem: t.Literal[False],
    ) -> dict[str, t.Any]: ...

    # FYI: this get_jwk method is duplicated in AuthLoginBaseClient
    # if this code is modified, please update that copy as well
    # this will ideally be resolved in a future SDK version by making this the only copy
    def get_jwk(
        self,
        openid_configuration: (
            GlobusHTTPResponse | dict[str, t.Any] | MissingType
        ) = MISSING,
        *,
        as_pem: bool = False,
    ) -> RSAPublicKey | dict[str, t.Any]:
        """
        Fetch the Globus Auth JWK.

        Returns either a dict or an RSA Public Key object depending on ``as_pem``.

        :param openid_configuration: The OIDC config as a GlobusHTTPResponse or dict.
            When not provided, it will be fetched automatically.
        :type openid_configuration: None | GlobusHTTPResponse | dict[str, typing.Any]
        :param as_pem: Decode the JWK to an RSA PEM key, typically for JWT decoding
        :type as_pem: bool
        """
        if openid_configuration is MISSING:
            log.debug("No OIDC Config provided, autofetching...")
            openid_configuration = self.get_openid_configuration()
        jwk_data = get_jwk_data(
            fget=self.get, openid_configuration=openid_configuration
        )
        return pem_decode_jwk_data(jwk_data=jwk_data) if as_pem else jwk_data

    def userinfo(self) -> GlobusHTTPResponse:
        """
        Call the Userinfo endpoint of Globus Auth.
        Userinfo is specified as part of the OpenID Connect (OIDC) standard,
        and Globus Auth's Userinfo is OIDC-compliant.

        The exact data returned will depend upon the set of OIDC-related scopes
        which were used to acquire the token being used for this call. For
        details, see the **API Info** below.

        .. tab-set::

            .. tab-item:: Example Usage

                .. code-block:: python

                    ac = AuthClient(...)
                    info = ac.userinfo()
                    print(
                        'Effective Identity "{info["sub"]}" has '
                        f'Full Name "{info["name"]}" and '
                        f'Email "{info["email"]}"'
                    )

            .. tab-item:: API Info

                ``GET /v2/oauth2/userinfo``

                .. extdoclink:: Get Userinfo
                    :ref: auth/reference/#get_or_post_v2_oauth2_userinfo_resource
        """
        log.debug("Looking up OIDC-style Userinfo from Globus Auth")
        return self.get("/v2/oauth2/userinfo")

    def get_identities(
        self,
        *,
        usernames: t.Iterable[str] | str | MissingType = MISSING,
        ids: t.Iterable[uuid.UUID | str] | uuid.UUID | str | MissingType = MISSING,
        provision: bool = False,
        query_params: dict[str, t.Any] | None = None,
    ) -> GetIdentitiesResponse:
        r"""
        Given ``usernames=<U>`` or (exclusive) ``ids=<I>`` as keyword
        arguments, looks up identity information for the set of identities
        provided.
        ``<U>`` and ``<I>`` in this case are comma-delimited strings listing
        multiple Identity Usernames or Identity IDs, or iterables of strings,
        each of which is an Identity Username or Identity ID.

        If Globus Auth's identity auto-provisioning behavior is desired,
        ``provision=True`` may be specified.

        Available with any authentication/client type.

        :param usernames: A username or list of usernames to lookup. Mutually exclusive
            with ``ids``
        :param ids: An identity ID or list of IDs to lookup. Mutually exclusive
            with ``usernames``
        :param provision: Create identities if they do not exist, allowing clients to
            get username-to-identity mappings prior to the identity being used
        :param query_params: Any additional parameters to be passed through
            as query params.

        .. tab-set::

            .. tab-item:: Example Usage

                .. code-block:: pycon

                    >>> ac = globus_sdk.AuthClient(...)
                    >>> # get by ID
                    >>> r = ac.get_identities(ids="46bd0f56-e24f-11e5-a510-131bef46955c")
                    >>> r.data
                    {
                      'identities': [
                        {
                          'email': None,
                          'id': '46bd0f56-e24f-11e5-a510-131bef46955c',
                          'identity_provider': '7daddf46-70c5-45ee-9f0f-7244fe7c8707',
                          'name': None,
                          'organization': None,
                          'status': 'unused',
                          'username': 'globus@globus.org'
                        }
                      ]
                    }
                    >>> ac.get_identities(
                    ...     ids=",".join(
                    ...         ("46bd0f56-e24f-11e5-a510-131bef46955c", "168edc3d-c6ba-478c-9cf8-541ff5ebdc1c")
                    ...     )
                    ... )
                    >>> # or by usernames
                    >>> ac.get_identities(usernames="globus@globus.org")
                    >>> ac.get_identities(usernames="globus@globus.org,auth@globus.org")

                You could also use iterables:

                .. code-block:: python

                    ac.get_identities(usernames=["globus@globus.org", "auth@globus.org"])

                    ac.get_identities(
                        ids=["46bd0f56-e24f-11e5-a510-131bef46955c", "168edc3d-c6ba-478c-9cf8-541ff5ebdc1c"]
                    )

                The result itself is iterable, so you can use it like so:

                .. code-block:: python

                    for identity in ac.get_identities(usernames=["globus@globus.org", "auth@globus.org"]):
                        print(identity["id"])

            .. tab-item:: API Info

                ``GET /v2/api/identities``

                .. extdoclink:: Get Identities
                    :ref: auth/reference/#v2_api_identities_resources
        """  # noqa: E501

        log.debug("Looking up Globus Auth Identities")

        query_params = {
            "usernames": commajoin(usernames),
            "ids": commajoin(ids),
            # only specify `provision` if `usernames` is given
            "provision": (
                str(provision).lower() if usernames is not MISSING else MISSING
            ),
            **(query_params or {}),
        }

        if (
            query_params["usernames"] is not MISSING
            and query_params["ids"] is not MISSING
        ):
            log.warning(
                "get_identities called with both usernames and "
                "identities set! Expecting an error."
            )

        log.debug(f"query_params={query_params}")

        return GetIdentitiesResponse(
            self.get("/v2/api/identities", query_params=query_params)
        )

    def get_identity_providers(
        self,
        *,
        domains: t.Iterable[str] | str | MissingType = MISSING,
        ids: t.Iterable[uuid.UUID | str] | uuid.UUID | str | MissingType = MISSING,
        query_params: dict[str, t.Any] | None = None,
    ) -> GetIdentityProvidersResponse:
        r"""
        Look up information about identity providers by domains or by IDs.

        :param domains: A domain or iterable of domains to lookup. Mutually exclusive
            with ``ids``.
        :param ids: An identity provider ID or iterable of IDs to lookup. Mutually exclusive
            with ``domains``.
        :param query_params: Any additional parameters to be passed through
            as query params.

        .. tab-set::

            .. tab-item:: Example Usage

                .. code-block:: pycon

                    >>> ac = globus_sdk.AuthClient(...)
                    >>> # get by ID
                    >>> r = ac.get_identity_providers(ids="41143743-f3c8-4d60-bbdb-eeecaba85bd9")
                    >>> r.data
                    {
                      'identity_providers': [
                        {
                          'alternative_names': [],
                          'name': 'Globus ID',
                          'domains': ['globusid.org'],
                          'id': '41143743-f3c8-4d60-bbdb-eeecaba85bd9',
                          'short_name': 'globusid'
                        }
                      ]
                    }
                    >>> ac.get_identities(
                    ...     ids=["41143743-f3c8-4d60-bbdb-eeecaba85bd9", "927d7238-f917-4eb2-9ace-c523fa9ba34e"]
                    ... )
                    >>> # or by domain
                    >>> ac.get_identities(domains="globusid.org")
                    >>> ac.get_identities(domains=["globus.org", "globusid.org"])

                The result itself is iterable, so you can use it like so:

                .. code-block:: python

                    for idp in ac.get_identity_providers(domains=["globus.org", "globusid.org"]):
                        print(f"name: {idp['name']}")
                        print(f"id: {idp['id']}")
                        print(f"domains: {idp['domains']}")
                        print()

            .. tab-item:: Example Response Data

                .. expandtestfixture:: auth.get_identity_providers

            .. tab-item:: API Info

                ``GET /v2/api/identity_providers``

                .. extdoclink:: Get Identity Providers
                    :ref: auth/reference/#get_identity_providers
        """  # noqa: E501

        log.debug("Looking up Globus Auth Identity Providers")

        if domains is not MISSING and ids is not MISSING:
            raise exc.GlobusSDKUsageError(
                "AuthClient.get_identity_providers does not take both "
                "'domains' and 'ids'. These are mutually exclusive."
            )
        elif domains is MISSING and ids is MISSING:
            log.warning(
                "Neither 'domains' nor 'ids' provided to get_identity_providers(). "
                "This can only succeed if 'query_params' were given."
            )

        query_params = {
            "domains": commajoin(domains),
            "ids": commajoin(ids),
            **(query_params or {}),
        }

        log.debug(f"query_params={query_params}")
        return GetIdentityProvidersResponse(
            self.get("/v2/api/identity_providers", query_params=query_params)
        )

    #
    # Developer APIs
    #

    def get_project(self, project_id: uuid.UUID | str) -> GlobusHTTPResponse:
        """
        Look up a project. Requires the ``manage_projects`` scope.

        :param project_id: The ID of the project to lookup

        .. tab-set::

            .. tab-item:: Example Usage

                .. code-block:: pycon

                    >>> ac = globus_sdk.AuthClient(...)
                    >>> r = ac.get_project("927d7238-f917-4eb2-9ace-c523fa9ba34e")
                    >>> r.data
                    {
                      'project': {
                        'admin_ids': ['41143743-f3c8-4d60-bbdb-eeecaba85bd9']
                        'contact_email': 'support@globus.org',
                        'display_name': 'Globus SDK Demo Project',
                        'admin_group_ids': None,
                        'id': '927d7238-f917-4eb2-9ace-c523fa9ba34e',
                        'project_name': 'Globus SDK Demo Project',
                        'admins': {
                          'identities': ['41143743-f3c8-4d60-bbdb-eeecaba85bd9'],
                          'groups': [],
                        },
                      }
                    }

            .. tab-item:: Example Response Data

                .. expandtestfixture:: auth.get_project

            .. tab-item:: API Info

                ``GET /v2/api/projects/{project_id}``

                .. extdoclink:: Get Projects
                    :ref: auth/reference/#get_projects
        """
        return self.get(f"/v2/api/projects/{project_id}")

    def get_projects(self) -> IterableResponse:
        """
        Look up projects on which the authenticated user is an admin.
        Requires the ``manage_projects`` scope.

        .. tab-set::

            .. tab-item:: Example Usage

                .. code-block:: pycon

                    >>> ac = globus_sdk.AuthClient(...)
                    >>> r = ac.get_projects()
                    >>> r.data
                    {
                      'projects': [
                        {
                          'admin_ids': ['41143743-f3c8-4d60-bbdb-eeecaba85bd9']
                          'contact_email': 'support@globus.org',
                          'display_name': 'Globus SDK Demo Project',
                          'admin_group_ids': None,
                          'id': '927d7238-f917-4eb2-9ace-c523fa9ba34e',
                          'project_name': 'Globus SDK Demo Project',
                          'admins': {
                            'identities': ['41143743-f3c8-4d60-bbdb-eeecaba85bd9'],
                            'groups': [],
                          },
                        }
                      ]
                    }

                The result itself is iterable, so you can use it like so:

                .. code-block:: python

                    for project in ac.get_projects():
                        print(f"name: {project['display_name']}")
                        print(f"id: {project['id']}")
                        print()

            .. tab-item:: Example Response Data

                .. expandtestfixture:: auth.get_projects

            .. tab-item:: API Info

                ``GET /v2/api/projects``

                .. extdoclink:: Get Projects
                    :ref: auth/reference/#get_projects
        """  # noqa: E501
        return GetProjectsResponse(self.get("/v2/api/projects"))

    def create_project(
        self,
        display_name: str,
        contact_email: str,
        *,
        admin_ids: (
            uuid.UUID | str | t.Iterable[uuid.UUID | str] | MissingType
        ) = MISSING,
        admin_group_ids: (
            uuid.UUID | str | t.Iterable[uuid.UUID | str] | MissingType
        ) = MISSING,
    ) -> GlobusHTTPResponse:
        """
        Create a new project. Requires the ``manage_projects`` scope.

        At least one of ``admin_ids`` or ``admin_group_ids`` must be provided.

        :param display_name: The name of the project
        :param contact_email: The email address of the project's point of contact
        :param admin_ids: A list of user IDs to be added as admins of the project
        :param admin_group_ids: A list of group IDs to be added as admins of the project

        .. tab-set::

            .. tab-item:: Example Usage

                When creating a project, your account is not necessarily included as an
                admin. The following snippet uses the ``manage_projects`` scope as well
                as the ``openid`` and ``email`` scopes to get the current user ID and
                email address and use those data to setup the project.

                .. code-block:: pycon

                    >>> ac = globus_sdk.AuthClient(...)
                    >>> userinfo = ac.userinfo()
                    >>> identity_id = userinfo["sub"]
                    >>> email = userinfo["email"]
                    >>> r = ac.create_project(
                    ...     "My Project",
                    ...     contact_email=email,
                    ...     admin_ids=identity_id,
                    ... )
                    >>> project_id = r["project"]["id"]

            .. tab-item:: Example Response Data

                .. expandtestfixture:: auth.create_project

            .. tab-item:: API Info

                ``POST /v2/api/projects``

                .. extdoclink:: Create Project
                    :ref: auth/reference/#create_project
        """
        body = {
            "display_name": display_name,
            "contact_email": contact_email,
            "admin_ids": strseq_listify(admin_ids),
            "admin_group_ids": strseq_listify(admin_group_ids),
        }
        return self.post("/v2/api/projects", data={"project": body})

    def update_project(
        self,
        project_id: uuid.UUID | str,
        *,
        display_name: str | MissingType = MISSING,
        contact_email: str | MissingType = MISSING,
        admin_ids: (
            uuid.UUID | str | t.Iterable[uuid.UUID | str] | MissingType
        ) = MISSING,
        admin_group_ids: (
            uuid.UUID | str | t.Iterable[uuid.UUID | str] | MissingType
        ) = MISSING,
    ) -> GlobusHTTPResponse:
        """
        Update a project. Requires the ``manage_projects`` scope.

        :param project_id: The ID of the project to update
        :param display_name: The name of the project
        :param contact_email: The email address of the project's point of contact
        :param admin_ids: A list of user IDs to be set as admins of the project
        :param admin_group_ids: A list of group IDs to be set as admins of the project

        .. tab-set::

            .. tab-item:: Example Usage

                The following snippet uses the ``manage_projects`` scope as well
                as the ``email`` scope to get the current user email address and set it
                as a project's contact email:

                .. code-block:: pycon

                    >>> ac = globus_sdk.AuthClient(...)
                    >>> project_id = ...
                    >>> userinfo = ac.userinfo()
                    >>> email = userinfo["email"]
                    >>> r = ac.update_project(project_id, contact_email=email)

            .. tab-item:: Example Response Data

                .. expandtestfixture:: auth.update_project

            .. tab-item:: API Info

                ``POST /v2/api/projects``

                .. extdoclink:: Update Project
                    :ref: auth/reference/#update_project
        """
        body = {
            "display_name": display_name,
            "contact_email": contact_email,
            "admin_ids": strseq_listify(admin_ids),
            "admin_group_ids": strseq_listify(admin_group_ids),
        }
        return self.put(f"/v2/api/projects/{project_id}", data={"project": body})

    def delete_project(self, project_id: uuid.UUID | str) -> GlobusHTTPResponse:
        """
        Delete a project. Requires the ``manage_projects`` scope.

        :param project_id: The ID of the project to delete

        .. tab-set::

            .. tab-item:: Example Usage

                .. code-block:: pycon

                    >>> ac = globus_sdk.AuthClient(...)
                    >>> project_id = ...
                    >>> r = ac.delete_project(project_id)

            .. tab-item:: Example Response Data

                .. expandtestfixture:: auth.delete_project

            .. tab-item:: API Info

                ``DELETE /v2/api/projects/{project_id}``

                .. extdoclink:: Delete Project
                    :ref: auth/reference/#delete_project
        """
        return self.delete(f"/v2/api/projects/{project_id}")

    def get_policy(self, policy_id: uuid.UUID | str) -> GlobusHTTPResponse:
        """
        Look up a policy. Requires the ``manage_projects`` scope.

        :param policy_id: The ID of the policy to lookup

        .. tab-set::

            .. tab-item:: Example Usage

                .. code-block:: pycon

                    >>> ac = globus_sdk.AuthClient(...)
                    >>> r = ac.get_policy("f5eaae7e-807f-41be-891a-ec86ff88df8f")
                    >>> r.data
                    {
                      'policy': {
                        'high_assurance': False,
                        'domain_constraints_include': ['globus.org'],
                        'display_name': 'Display Name',
                        'description': 'Description',
                        'id': 'f5eaae7e-807f-41be-891a-ec86ff88df8f',
                        'domain_constraints_exclude': None,
                        'project_id': 'da84e531-1afb-43cb-8c87-135ab580516a',
                        'authentication_assurance_timeout': 35,
                        'required_mfa": False
                      }
                    }

            .. tab-item:: Example Response Data

                .. expandtestfixture:: auth.get_policy

            .. tab-item:: API Info

                ``GET /v2/api/policies/{policy_id}``

                .. extdoclink:: Get Policies
                    :ref: auth/reference/#get_policies
        """
        return self.get(f"/v2/api/policies/{policy_id}")

    def get_policies(self) -> IterableResponse:
        """
        Look up policies in projects on which the authenticated user is an admin.
        Requires the ``manage_projects`` scope.

        .. tab-set::

            .. tab-item:: Example Usage

                .. code-block:: pycon

                    >>> ac = globus_sdk.AuthClient(...)
                    >>> r = ac.get_policies()
                    >>> r.data
                    {
                      'policies': [
                        {
                          'high_assurance': False,
                          'domain_constraints_include': ['greenlight.org'],
                          'display_name': 'GreenLight domain Only Policy',
                          'description': 'Only allow access from @greenlight.org',
                          'id': '99d2dc75-3acb-48ff-b5e5-2eee0a5121d1',
                          'domain_constraints_exclude': None,
                          'project_id': 'da84e531-1afb-43cb-8c87-135ab580516a',
                          'authentication_assurance_timeout': 35,
                          'required_mfa': False
                        },
                        {
                          'high_assurance': True,
                          'domain_constraints_include': None,
                          'display_name': 'No RedLight domain Policy',
                          'description': 'Disallow access from @redlight.org',
                          'id': '5d93ebf0-b4c6-4928-9929-4ac47fc2786d',
                          'domain_constraints_exclude': ['redlight.org'],
                          'project_id': 'da84e531-1afb-43cb-8c87-135ab580516a',
                          'authentication_assurance_timeout': 35,
                          'required_mfa': True
                        }
                      ]
                    }

            .. tab-item:: Example Response Data

                .. expandtestfixture:: auth.get_policies

            .. tab-item:: API Info

                ``GET /v2/api/policies``

                .. extdoclink:: Get Policies
                    :ref: auth/reference/#get_policies
        """
        return GetPoliciesResponse(self.get("/v2/api/policies"))

    def create_policy(
        self,
        *,
        project_id: uuid.UUID | str,
        display_name: str,
        description: str,
        high_assurance: bool | MissingType = MISSING,
        authentication_assurance_timeout: int | MissingType = MISSING,
        required_mfa: bool | MissingType = MISSING,
        domain_constraints_include: t.Iterable[str] | None | MissingType = MISSING,
        domain_constraints_exclude: t.Iterable[str] | None | MissingType = MISSING,
    ) -> GlobusHTTPResponse:
        """
        Create a new Auth policy. Requires the ``manage_projects`` scope.

        :param project_id: ID of the project for the new policy
        :param display_name: A user-friendly name for the policy
        :param description: A user-friendly description to explain the purpose of the
            policy
        :param high_assurance: Whether or not this policy is applied to sessions.
        :param authentication_assurance_timeout: Number of seconds within which someone
            must have authenticated to satisfy the policy
        :param required_mfa: If True, then multi-factor authentication is required.
            This can only be set to True for high-assurance policies. The default
            is False.
        :param domain_constraints_include: A list of domains that can satisfy the policy
        :param domain_constraints_exclude: A list of domains that cannot satisfy the
            policy

        .. note::

            ``project_id``, ``display_name``, and ``description`` are all required
            arguments, although they are not declared as required in the function
            signature. This is due to a backwards compatible behavior with earlier
            versions of globus-sdk, and will be changed in a future release which
            removes the compatible behavior.

        .. tab-set::

            .. tab-item:: Example Usage

                .. code-block:: pycon

                    >>> ac = globus_sdk.AuthClient(...)
                    >>> r = ac.create_policy(
                    ...     project_id="da84e531-1afb-43cb-8c87-135ab580516a",
                    ...     high_assurance=True,
                    ...     authentication_assurance_timeout=35,
                    ...     required_mfa=True,
                    ...     display_name="No RedLight domain Policy",
                    ...     description="Disallow access from @redlight.org",
                    ...     domain_constraints_exclude=["redlight.org"],
                    ... )
                    >>> policy_id = r["policy"]["id"]

            .. tab-item:: Example Response Data

                .. expandtestfixture:: auth.create_policy

            .. tab-item:: API Info

                ``POST /v2/api/policies``

                .. extdoclink:: Create Policy
                    :ref: auth/reference/#create_policy
        """
        body: dict[str, t.Any] = {
            "project_id": project_id,
            "high_assurance": high_assurance,
            "authentication_assurance_timeout": authentication_assurance_timeout,
            "required_mfa": required_mfa,
            "display_name": display_name,
            "description": description,
            "domain_constraints_include": strseq_listify(domain_constraints_include),
            "domain_constraints_exclude": strseq_listify(domain_constraints_exclude),
        }

        return self.post("/v2/api/policies", data={"policy": body})

    def update_policy(
        self,
        policy_id: uuid.UUID | str,
        *,
        project_id: uuid.UUID | str | MissingType = MISSING,
        authentication_assurance_timeout: int | MissingType = MISSING,
        required_mfa: bool | MissingType = MISSING,
        display_name: str | MissingType = MISSING,
        description: str | MissingType = MISSING,
        domain_constraints_include: t.Iterable[str] | None | MissingType = MISSING,
        domain_constraints_exclude: t.Iterable[str] | None | MissingType = MISSING,
    ) -> GlobusHTTPResponse:
        """
        Update a policy. Requires the ``manage_projects`` scope.

        :param policy_id: ID of the policy to update
        :param project_id: ID of the project for the new policy
        :param authentication_assurance_timeout: Number of seconds within which someone
            must have authenticated to satisfy the policy
        :param required_mfa: If True, then multi-factor authentication is required.
            This can only be set to True for high-assurance policies. The default
            is False.
        :param display_name: A user-friendly name for the policy
        :param description: A user-friendly description to explain the purpose of the
            policy
        :param domain_constraints_include: A list of domains that can satisfy the policy
        :param domain_constraints_exclude: A list of domains that can not satisfy the
            policy

        .. tab-set::

            .. tab-item:: Example Usage

                .. code-block:: pycon

                    >>> ac = globus_sdk.AuthClient(...)
                    >>> policy_id = ...
                    >>> r = ac.update_policy(scope_id, display_name="Greenlight Policy")

            .. tab-item:: Example Response Data

                .. expandtestfixture:: auth.update_policy

            .. tab-item:: API Info

                ``POST /v2/api/policies/{policy_id}``

                .. extdoclink:: Update Policy
                    :ref: auth/reference/#update_policy
        """
        body: dict[str, t.Any] = {
            "authentication_assurance_timeout": authentication_assurance_timeout,
            "required_mfa": required_mfa,
            "display_name": display_name,
            "description": description,
            "domain_constraints_include": strseq_listify(domain_constraints_include),
            "domain_constraints_exclude": strseq_listify(domain_constraints_exclude),
            "project_id": project_id,
        }
        return self.put(f"/v2/api/policies/{policy_id}", data={"policy": body})

    def delete_policy(self, policy_id: uuid.UUID | str) -> GlobusHTTPResponse:
        """
        Delete a policy. Requires the ``manage_projects`` scope.

        :param policy_id: The ID of the policy to delete

        .. tab-set::

            .. tab-item:: Example Usage

                .. code-block:: pycon

                    >>> ac = globus_sdk.AuthClient(...)
                    >>> policy_id = ...
                    >>> r = ac.delete_policy(policy_id)

            .. tab-item:: Example Response Data

                .. expandtestfixture:: auth.delete_policy

            .. tab-item:: API Info

                ``DELETE /v2/api/policies/{policy_id}``

                .. extdoclink:: Delete Policy
                    :ref: auth/reference/#delete_policy
        """
        return self.delete(f"/v2/api/policies/{policy_id}")

    def get_client(
        self,
        *,
        client_id: uuid.UUID | str | MissingType = MISSING,
        fqdn: str | MissingType = MISSING,
    ) -> GlobusHTTPResponse:
        """
        Look up a client by ``client_id`` or (exclusive) by ``fqdn``.
        Requires the ``manage_projects`` scope.

        :param client_id: The ID of the client to look up
        :param fqdn: The fully-qualified domain name of the client to look up

        .. tab-set::

            .. tab-item:: Example Usage

                .. code-block:: pycon

                    >>> ac = globus_sdk.AuthClient(...)
                    >>> # by client_id
                    >>> r = ac.get_client(client_id="6336437e-37e8-4559-82a8-674390c1fd2e")
                    >>> r.data
                    {
                      'client': {
                        'required_idp': None,
                        'name': 'Great client of FOO',
                        'redirect_uris': [],
                        'links': {
                          'privacy_policy': None,
                          'terms_and_conditions': None
                        },
                        'scopes': [],
                        'grant_types': [
                          'authorization_code',
                          'client_credentials',
                          'refresh_token'
                        ],
                        'id': '6336437e-37e8-4559-82a8-674390c1fd2e',
                        'prompt_for_named_grant': False,
                        'fqdns': ['globus.org'],
                        'project': 'da84e531-1afb-43cb-8c87-135ab580516a',
                        'client_type': 'client_identity',
                        'visibility': 'private',
                        'parent_client': None,
                        'userinfo_from_effective_identity': True,
                        'preselect_idp': None,
                        'public_client': False
                      }
                    }
                    >>> # by fqdn
                    >>> fqdn = ...
                    >>> r = ac.get_client(fqdn=fqdn)

            .. tab-item:: Example Response Data

                .. expandtestfixture:: auth.get_client

            .. tab-item:: API Info

                ``GET /v2/api/clients/{client_id}``
                ``GET /v2/api/clients?fqdn={fqdn}``

                .. extdoclink:: Get Clients
                    :ref: auth/reference/#get_clients
        """  # noqa: E501
        if client_id is not MISSING and fqdn is not MISSING:
            raise exc.GlobusSDKUsageError(
                "AuthClient.get_client does not take both "
                "'client_id' and 'fqdn'. These are mutually exclusive."
            )

        if client_id is MISSING and fqdn is MISSING:
            raise exc.GlobusSDKUsageError(
                "AuthClient.get_client requires either 'client_id' or 'fqdn'."
            )

        if client_id is not MISSING:
            return self.get(f"/v2/api/clients/{client_id}")
        return self.get("/v2/api/clients", query_params={"fqdn": fqdn})

    def get_clients(self) -> IterableResponse:
        """
        Look up clients in projects on which the authenticated user is an admin.
        Requires the ``manage_projects`` scope.

        .. tab-set::

            .. tab-item:: Example Usage

                .. code-block:: pycon

                    >>> ac = globus_sdk.AuthClient(...)
                    >>> r = ac.get_clients()
                    >>> r.data
                    {
                      'clients': [
                        {
                          'required_idp': None,
                          'name': 'Great client of FOO',
                          'redirect_uris': [],
                          'links': {'privacy_policy': None, 'terms_and_conditions': None},
                          'scopes': [],
                          'grant_types': ['authorization_code', 'client_credentials', 'refresh_token'],
                          'id': 'b6001d11-8765-49d3-a503-ba323fc74eee',
                          'prompt_for_named_grant': False,
                          'fqdns': ['foo.net'],
                          'project': 'da84e531-1afb-43cb-8c87-135ab580516a',
                          'client_type': 'client_identity',
                          'visibility': 'private',
                          'parent_client': None,
                          'userinfo_from_effective_identity': True,
                          'preselect_idp': None,
                          'public_client': False,
                        },
                        {
                          'required_idp': None,
                          'name': 'Lessor client of BAR',
                          'redirect_uris': [],
                          'links': {'privacy_policy': None, 'terms_and_conditions': None},
                          'scopes': [],
                          'grant_types': ['authorization_code', 'client_credentials', 'refresh_token'],
                          'id': 'b87f7415-ddf9-4868-8e55-d10c065f733d',
                          'prompt_for_named_grant': False,
                          'fqdns': ['bar.org'],
                          'project': 'da84e531-1afb-43cb-8c87-135ab580516a',
                          'client_type': 'client_identity',
                          'visibility': 'private',
                          'parent_client': None,
                          'userinfo_from_effective_identity': True,
                          'preselect_idp': None,
                          'public_client': False,
                        }
                      ]
                    }

            .. tab-item:: Example Response Data

                .. expandtestfixture:: auth.get_clients

            .. tab-item:: API Info

                ``GET /v2/api/clients``

                .. extdoclink:: Get Clients
                    :ref: auth/reference/#get_clients
        """  # noqa: E501
        return GetClientsResponse(self.get("/v2/api/clients"))

    def create_client(
        self,
        name: str,
        project: uuid.UUID | str,
        *,
        public_client: bool | MissingType = MISSING,
        client_type: (
            MissingType
            | t.Literal[
                "client_identity",
                "confidential_client",
                "globus_connect_server",
                "public_installed_client",
                "hybrid_confidential_client_resource_server",
                "resource_server",
            ]
        ) = MISSING,
        visibility: MissingType | t.Literal["public", "private"] = MISSING,
        redirect_uris: t.Iterable[str] | MissingType = MISSING,
        terms_and_conditions: str | MissingType = MISSING,
        privacy_policy: str | MissingType = MISSING,
        required_idp: uuid.UUID | str | MissingType = MISSING,
        preselect_idp: uuid.UUID | str | MissingType = MISSING,
        additional_fields: dict[str, t.Any] | MissingType = MISSING,
    ) -> GlobusHTTPResponse:
        """
        Create a new client. Requires the ``manage_projects`` scope.

        :param name: The display name shown to users on consents. May not contain
            linebreaks.
        :param project: ID representing the project this client belongs to.

        :param public_client: This is used to infer which OAuth grant_types the client
            will be able to use. Should be false if the client is capable of keeping
            secret credentials (such as clients running on a server) and true if it is
            not (such as native apps). After creation this value is immutable. This
            option is mutually exclusive with ``client_type``, exactly one must be
            given.
        :param client_type: Defines the type of client that will be created. This
            option is mutually exclusive with ``public_client``, exactly one must
            be given.

            .. dropdown:: Values for ``client_type``

                .. list-table::

                    * - ``"confidential_client"``
                      - Applications that are OAuth confidential clients, and can
                        manage a client secret and requests for user consent.
                    * - ``"public_installed_client"``
                      - Applications that are OAuth public clients or native
                        applications that are distributed to users, and thus cannot
                        manage a client secret.
                    * - ``"client_identity"``
                      - Applications that authenticate and act as the application
                        itself. These applications are used for automation and as
                        service or community accounts, and do NOT act on behalf of
                        other users. Also known as a "Service Account".
                    * - ``"resource_server"``
                      - An API (OAuth resource server) that uses Globus Auth tokens for
                        authentication. Users accessing the service login via Globus and
                        consent for the client to use your API.
                    * - ``"globus_connect_server"``
                      - Create a client that will service as a Globus Connect Server
                        endpoint.
                    * - ``"hybrid_confidential_client_resource_server"``
                      - A client which can use any behavior with Globus Auth - an
                        application (confidential or public client), service account,
                        or API.

        :param visibility: If set to "public", any authenticated entity can view it.
            When set to "private", only entities in the same project as the client can
            view it.
        :param redirect_uris: list of URIs that may be used in OAuth authorization
            flows.
        :param terms_and_conditions: URL of client's terms and conditions.
        :param privacy_policy: URL of client's privacy policy.
        :param required_idp: In order to use this client a user must have an identity
            from this IdP in their identity set.
        :param preselect_idp: This preselects the given IdP on the Globus Auth login
            page if the user is not already authenticated.
        :param additional_fields: Any additional parameters to be passed through.

        .. tab-set::

            .. tab-item:: Example Usage

                .. code-block:: pycon

                    >>> ac = globus_sdk.AuthClient(...)
                    >>> project = ...
                    >>> r = ac.create_client(
                    ...     "My client",
                    ...     True,
                    ...     project,
                    ...     True,
                    ... )
                    >>> client_id = r["client"]["id"]

            .. tab-item:: Example Response Data

                .. expandtestfixture:: auth.create_client

            .. tab-item:: API Info

                ``POST /v2/api/clients``

                .. extdoclink:: Create Client
                    :ref: auth/reference/#create_client
        """
        # Must specify exactly one of public_client or client_type
        if public_client is not MISSING and client_type is not MISSING:
            raise exc.GlobusSDKUsageError(
                "AuthClient.create_client does not take both "
                "'public_client' and 'client_type'. These are mutually exclusive."
            )
        if public_client is MISSING and client_type is MISSING:
            raise exc.GlobusSDKUsageError(
                "AuthClient.create_client requires either 'public_client' or "
                "'client_type'."
            )

        body: dict[str, t.Any] = {
            "name": name,
            "project": project,
            "visibility": visibility,
            "redirect_uris": redirect_uris,
            "required_idp": required_idp,
            "preselect_idp": preselect_idp,
            "public_client": public_client,
            "client_type": client_type,
        }
        # terms_and_conditions and privacy_policy must both be set or unset
        if bool(terms_and_conditions) ^ bool(privacy_policy):
            raise exc.GlobusSDKUsageError(
                "terms_and_conditions and privacy_policy must both be set or unset"
            )
        links: dict[str, str | MissingType] = {
            "terms_and_conditions": terms_and_conditions,
            "privacy_policy": privacy_policy,
        }
        if terms_and_conditions or privacy_policy:
            body["links"] = links

        if additional_fields is not MISSING:
            body.update(additional_fields)

        return self.post("/v2/api/clients", data={"client": body})

    def update_client(
        self,
        client_id: uuid.UUID | str,
        *,
        name: str | MissingType = MISSING,
        visibility: MissingType | t.Literal["public", "private"] = MISSING,
        redirect_uris: t.Iterable[str] | MissingType = MISSING,
        terms_and_conditions: str | None | MissingType = MISSING,
        privacy_policy: str | None | MissingType = MISSING,
        required_idp: uuid.UUID | str | None | MissingType = MISSING,
        preselect_idp: uuid.UUID | str | None | MissingType = MISSING,
        additional_fields: dict[str, t.Any] | MissingType = MISSING,
    ) -> GlobusHTTPResponse:
        """
        Update a client. Requires the ``manage_projects`` scope.

        :param client_id: ID of the client to update
        :param name: The display name shown to users on consents. May not contain
            linebreaks.
        :param visibility: If set to "public", any authenticated entity can view it.
            When set to "private", only entities in the same project as the client can
            view it.
        :param redirect_uris: list of URIs that may be used in OAuth authorization
            flows.
        :param terms_and_conditions: URL of client's terms and conditions.
        :param privacy_policy: URL of client's privacy policy.
        :param required_idp: In order to use this client a user must have an identity
            from this IdP in their identity set.
        :param preselect_idp: This preselects the given IdP on the Globus Auth login
            page if the user is not already authenticated.
        :param additional_fields: Any additional parameters to be passed through.


        .. tab-set::

            .. tab-item:: Example Usage

                .. code-block:: pycon

                    >>> ac = globus_sdk.AuthClient(...)
                    >>> client_id = ...
                    >>> r = ac.update_client(client_id, name="Foo Utility")

            .. tab-item:: Example Response Data

                .. expandtestfixture:: auth.update_client

            .. tab-item:: API Info

                ``POST /v2/api/clients/{client_id}``

                .. extdoclink:: Update Client
                    :ref: auth/reference/#update_client
        """
        body: dict[str, t.Any] = {
            "name": name,
            "visibility": visibility,
            "redirect_uris": redirect_uris,
            "required_idp": required_idp,
            "preselect_idp": preselect_idp,
        }

        # terms_and_conditions and privacy_policy must both be set or unset, and if one
        # is set to `None` they both must be set to `None`
        # note the subtle differences between this logic for "update" and the matching
        # logic for "create"
        # "create" does not need to handle `None` as a distinct and meaningful value
        if type(terms_and_conditions) is not type(privacy_policy):
            raise exc.GlobusSDKUsageError(
                "terms_and_conditions and privacy_policy must both be set or unset"
            )
        links: dict[str, str | None | MissingType] = {
            "terms_and_conditions": terms_and_conditions,
            "privacy_policy": privacy_policy,
        }
        if terms_and_conditions is not MISSING or privacy_policy is not MISSING:
            body["links"] = links

        if additional_fields is not MISSING:
            body.update(additional_fields)

        return self.put(f"/v2/api/clients/{client_id}", data={"client": body})

    def delete_client(self, client_id: uuid.UUID | str) -> GlobusHTTPResponse:
        """
        Delete a client. Requires the ``manage_projects`` scope.

        :param client_id: The ID of the client to delete

        .. tab-set::

            .. tab-item:: Example Usage

                .. code-block:: pycon

                    >>> ac = globus_sdk.AuthClient(...)
                    >>> client_id = ...
                    >>> r = ac.delete_policy(client_id)

            .. tab-item:: Example Response Data

                .. expandtestfixture:: auth.delete_client

            .. tab-item:: API Info

                ``DELETE /v2/api/clients/{client_id}``

                .. extdoclink:: Delete Client
                    :ref: auth/reference/#delete_client
        """
        return self.delete(f"/v2/api/clients/{client_id}")

    def get_client_credentials(self, client_id: uuid.UUID | str) -> IterableResponse:
        """
        Look up client credentials by ``client_id``.  Requires the
        ``manage_projects`` scope.

        :param client_id: The ID of the client that owns the credentials

        .. tab-set::

            .. tab-item:: Example Usage

                .. code-block:: pycon

                    >>> ac = globus_sdk.AuthClient(...)
                    >>> r = ac.get_credentials("6336437e-37e8-4559-82a8-674390c1fd2e")
                    >>> r.data
                    {
                      'credentials': [
                        'name': 'foo',
                        'id': 'cf88318e-b2dd-43fd-8ea5-2086fc69ffac',
                        'created': '2023-10-21T22:46:15.845937+00:00',
                        'client': '6336437e-37e8-4559-82a8-674390c1fd2e',
                        'secret': None,
                      ]
                    }

            .. tab-item:: Example Response Data

                .. expandtestfixture:: auth.get_client_credentials

            .. tab-item:: API Info

                ``GET /v2/api/clients/{client_id}/credentials``

                .. extdoclink:: Get Client Credentials
                    :ref: auth/reference/#get_client_credentials
        """  # noqa: E501
        return GetClientCredentialsResponse(
            self.get(f"/v2/api/clients/{client_id}/credentials")
        )

    def create_client_credential(
        self,
        client_id: uuid.UUID | str,
        name: str,
    ) -> GlobusHTTPResponse:
        """
        Create a new client credential. Requires the ``manage_projects`` scope.

        :param client_id: ID for the client
        :param name: The display name of the new credential.

        .. tab-set::

            .. tab-item:: Example Usage

                .. code-block:: pycon

                    >>> ac = globus_sdk.AuthClient(...)
                    >>> client_id = ...
                    >>> name = ...
                    >>> r = ac.create_client_credential(
                    ...     "25afc56d-02af-4175-8c90-9941ebb623dd",
                    ...     "New Credentials",
                    ... )
                    >>> r.data
                    {
                        'name': 'New Credentials',
                        'id': '3a53cb4d-edd6-4ae3-900e-25b38b5fce02',
                        'created': '2023-10-21T22:46:15.845937+00:00',
                        'client': '25afc56d-02af-4175-8c90-9941ebb623dd',
                        'secret': 'abc123',
                    }

            .. tab-item:: Example Response Data

                .. expandtestfixture:: auth.create_client_credential

            .. tab-item:: API Info

                ``POST /v2/api/clients/{client_id}/credentials``

                .. extdoclink:: Create Client Credentials
                    :ref: auth/reference/#create_client_credential
        """
        return self.post(
            f"/v2/api/clients/{client_id}/credentials",
            data={"credential": {"name": name}},
        )

    def delete_client_credential(
        self,
        client_id: uuid.UUID | str,
        credential_id: uuid.UUID | str,
    ) -> GlobusHTTPResponse:
        """
        Delete a credential. Requires the ``manage_projects`` scope.

        :param client_id: The ID of the client that owns the credential to delete
        :param credential_id: The ID of the credential to delete

        .. tab-set::

            .. tab-item:: Example Usage

                .. code-block:: pycon

                    >>> ac = globus_sdk.AuthClient(...)
                    >>> client_id = ...
                    >>> credential_id = ...
                    >>> r = ac.delete_policy(client_id, credential_id)

            .. tab-item:: Example Response Data

                .. expandtestfixture:: auth.delete_client_credential

            .. tab-item:: API Info

                ``DELETE /v2/api/clients/{client_id}/credentials/{credential_id}``

                .. extdoclink:: Delete Credential
                    :ref: auth/reference/#delete_client_credentials
        """
        return self.delete(f"/v2/api/clients/{client_id}/credentials/{credential_id}")

    def get_scope(self, scope_id: uuid.UUID | str) -> GlobusHTTPResponse:
        """
        Look up a scope by ``scope_id``.  Requires the ``manage_projects`` scope.

        :param scope_id: The ID of the scope to look up

        .. tab-set::

            .. tab-item:: Example Usage

                .. code-block:: pycon

                    >>> ac = globus_sdk.AuthClient(...)
                    >>> r = ac.get_scope(scope_id="6336437e-37e8-4559-82a8-674390c1fd2e")
                    >>> r.data
                    {
                      'scope': {
                        'scope_string': 'https://auth.globus.org/scopes/3f33d83f-ec0a-4190-887d-0622e7c4ee9a/manager',
                        'allows_refresh_token': False,
                        'id': '87cf7b34-e1e1-4805-a8d5-51ab59fe6000',
                        'advertised': False,
                        'required_domains': [],
                        'name': 'Client manage scope',
                        'description': 'Manage configuration of this client',
                        'client': '3f33d83f-ec0a-4190-887d-0622e7c4ee9a',
                        'dependent_scopes': [],
                      }
                    }

            .. tab-item:: Example Response Data

                .. expandtestfixture:: auth.get_scope

            .. tab-item:: API Info

                ``GET /v2/api/scopes/{scope_id}``

                .. extdoclink:: Get Scopes
                    :ref: auth/reference/#get_scopes
        """  # noqa: E501
        return self.get(f"/v2/api/scopes/{scope_id}")

    def get_scopes(
        self,
        *,
        scope_strings: t.Iterable[str] | str | MissingType = MISSING,
        ids: t.Iterable[uuid.UUID | str] | uuid.UUID | str | MissingType = MISSING,
        query_params: dict[str, t.Any] | MissingType = MISSING,
    ) -> IterableResponse:
        """
        Look up scopes in projects on which the authenticated user is an admin.
        The scopes response can be filted by ``scope_strings`` or (exclusive)
        ``ids``.  Requires the ``manage_projects`` scope.

        :param scope_strings: The scope_strings of the scopes to look up
        :param ids: The ID of the scopes to look up
        :param query_params: Any additional parameters to be passed through
            as query params.

        .. tab-set::

            .. tab-item:: Example Usage

                .. code-block:: pycon

                    >>> ac = globus_sdk.AuthClient(...)
                    >>> # get all scopes
                    >>> r = ac.get_scopes()
                    >>> r.data
                    {
                      'scopes': [
                        {
                          'scope_string': 'https://auth.globus.org/scopes/3f33d83f-ec0a-4190-887d-0622e7c4ee9a/manage',
                          'allows_refresh_token': False,
                          'id': '70147193-f88a-4da9-9d6e-677c15e790e5',
                          'advertised': False,
                          'required_domains': [],
                          'name': 'Client manage scope',
                          'description': 'Manage configuration of this client',
                          'client': '3f33d83f-ec0a-4190-887d-0622e7c4ee9a',
                          'dependent_scopes': [],
                        },
                        {
                          'scope_string': 'https://auth.globus.org/scopes/dfc9a6d3-3373-4a6d-b0a1-b7026d1559d6/view',
                          'allows_refresh_token': False,
                          'id': '3793042a-203c-4e86-8dfe-17d407d0bb5f',
                          'advertised': False,
                          'required_domains': [],
                          'name': 'Client view scope',
                          'description': 'View configuration of this client',
                          'client': 'dfc9a6d3-3373-4a6d-b0a1-b7026d1559d6',
                          'dependent_scopes': [],
                        }
                      ]
                    }

                    >>> # by all scope ids
                    >>> scope_ids = ...
                    >>> r = ac.get_scopes(ids=scopes_ides)

                    >>> # by all scope strings
                    >>> scope_strings = ...
                    >>> r = ac.get_scopes(scope_strings=scope_strings)

            .. tab-item:: Example Response Data

                .. expandtestfixture:: auth.get_scopes

            .. tab-item:: API Info

                ``GET /v2/api/scopes``
                ``GET /v2/api/scopes?ids=...``
                ``GET /v2/api/scopes?scope_strings=...``

                .. extdoclink:: Get Scopes
                    :ref: auth/reference/#get_scopes
        """  # noqa: E501
        if scope_strings is not MISSING and ids is not MISSING:
            raise exc.GlobusSDKUsageError(
                "AuthClient.get_scopes does not take both "
                "'scopes_strings' and 'ids'. These are mutually exclusive."
            )

        if query_params is MISSING:
            query_params = {}

        if scope_strings is not MISSING:
            query_params["scope_strings"] = commajoin(scope_strings)
        if ids is not MISSING:
            query_params["ids"] = commajoin(ids)

        return GetScopesResponse(self.get("/v2/api/scopes", query_params=query_params))

    def create_scope(
        self,
        client_id: uuid.UUID | str,
        name: str,
        description: str,
        scope_suffix: str,
        *,
        required_domains: t.Iterable[str] | MissingType = MISSING,
        dependent_scopes: t.Iterable[DependentScopeSpec] | MissingType = MISSING,
        advertised: bool | MissingType = MISSING,
        allows_refresh_token: bool | MissingType = MISSING,
    ) -> GlobusHTTPResponse:
        """
        Create a new scope. Requires the ``manage_projects`` scope.

        :param client_id: ID of the client for the new scope
        :param name: A display name used to display consents to users,
            along with description
        :param description: A description used to display consents to users, along with
            name
        :param scope_suffix: String consisting of lowercase letters, number, and
            underscores. This will be the final part of the scope_string
        :param required_domains: Domains the user must have linked identities in in
            order to make use of the scope
        :param dependent_scopes: Scopes included in the consent for this new scope
        :param advertised: If True, scope is visible to anyone regardless of client
            visibility, otherwise, scope visibility is based on client visibility.
        :param allows_refresh_token: Whether or not the scope allows refresh tokens
            to be issued.

        .. tab-set::

            .. tab-item:: Example Usage

                .. code-block:: pycon

                    >>> ac = globus_sdk.AuthClient(...)
                    >>> client_id = ...
                    >>> r = ac.create_scope(
                    ...     client_id,
                    ...     "Client Management",
                    ...     "Manage client configuration",
                    ...     "manage",
                    ... )
                    >>> scope_id = r["scope"]["id"]

            .. tab-item:: Example Response Data

                .. expandtestfixture:: auth.create_scope

            .. tab-item:: API Info

                ``POST /v2/api/clients/{client_id}/scopes``

                .. extdoclink:: Create Scope
                    :ref: auth/reference/#create_scope
        """
        body: dict[str, t.Any] = {
            "name": name,
            "description": description,
            "scope_suffix": scope_suffix,
            "advertised": advertised,
            "allows_refresh_token": allows_refresh_token,
            "required_domains": required_domains,
            "dependent_scopes": dependent_scopes,
        }

        return self.post(f"/v2/api/clients/{client_id}/scopes", data={"scope": body})

    def update_scope(
        self,
        scope_id: uuid.UUID | str,
        *,
        name: str | MissingType = MISSING,
        description: str | MissingType = MISSING,
        scope_suffix: str | MissingType = MISSING,
        required_domains: t.Iterable[str] | MissingType = MISSING,
        dependent_scopes: t.Iterable[DependentScopeSpec] | MissingType = MISSING,
        advertised: bool | MissingType = MISSING,
        allows_refresh_token: bool | MissingType = MISSING,
    ) -> GlobusHTTPResponse:
        """
        Update a scope. Requires the ``manage_projects`` scope.

        :param scope_id: ID of the scope to update
        :param name: A display name used to display consents to users,
            along with description
        :param description: A description used to display consents to users, along with
            name
        :param scope_suffix: String consisting of lowercase letters, number, and
            underscores. This will be the final part of the scope_string
        :param required_domains: Domains the user must have linked identities in in
            order to make use of the scope
        :param dependent_scopes: Scopes included in the consent for this new scope
        :param advertised: If True, scope is visible to anyone regardless of client
            visibility, otherwise, scope visibility is based on client visibility.
        :param allows_refresh_token: Whether or not the scope allows refresh tokens
            to be issued.

        .. tab-set::

            .. tab-item:: Example Usage

                .. code-block:: pycon

                    >>> ac = globus_sdk.AuthClient(...)
                    >>> scope_id = ...
                    >>> r = ac.update_scope(scope_id, scope_suffix="manage")

            .. tab-item:: Example Response Data

                .. expandtestfixture:: auth.update_scope

            .. tab-item:: API Info

                ``POST /v2/api/scopes/{scope_id}``

                .. extdoclink:: Update Scope
                    :ref: auth/reference/#update_scope
        """
        body: dict[str, t.Any] = {
            "name": name,
            "description": description,
            "scope_suffix": scope_suffix,
            "advertised": advertised,
            "allows_refresh_token": allows_refresh_token,
            "required_domains": required_domains,
            "dependent_scopes": dependent_scopes,
        }

        return self.put(f"/v2/api/scopes/{scope_id}", data={"scope": body})

    def delete_scope(self, scope_id: uuid.UUID | str) -> GlobusHTTPResponse:
        """
        Delete a scope. Requires the ``manage_projects`` scope.

        :param scope_id: The ID of the scope to delete

        .. tab-set::

            .. tab-item:: Example Usage

                .. code-block:: pycon

                    >>> ac = globus_sdk.AuthClient(...)
                    >>> scope_id = ...
                    >>> r = ac.delete_scope(scope_id)

            .. tab-item:: Example Response Data

                .. expandtestfixture:: auth.delete_scope

            .. tab-item:: API Info

                ``DELETE /v2/api/scopes/{scope_id}``

                .. extdoclink:: Delete Scopes
                    :ref: auth/reference/#delete_scope
        """
        return self.delete(f"/v2/api/scopes/{scope_id}")

    def get_consents(
        self,
        identity_id: uuid.UUID | str,
        *,
        # pylint: disable=redefined-builtin
        all: bool = False,
    ) -> GetConsentsResponse:
        """
        Look up consents for a user.

        If requesting "all" consents, the view_consents scope is required.

        :param identity_id: The ID of the identity to look up consents for
        :param all: If true, return all consents, including those that have
            been issued to other clients. If false, return only consents rooted at this
            client id for the requested identity. Most clients should pass False.

        .. tab-set::

            .. tab-item:: Example Usage

                .. code-block:: pycon

                    >>> ac = globus_sdk.AuthClient(...)
                    >>> identity_id = ...
                    >>> forest = ac.get_consents(identity_id).to_forest()

            .. tab-item:: Example Response Data

                .. expandtestfixture:: auth.get_consents

            .. tab-item:: API Info

                ``GET /v2/api/identities/{identity_id}/consents``
        """
        return GetConsentsResponse(
            self.get(
                f"/v2/api/identities/{identity_id}/consents", query_params={"all": all}
            )
        )



#!/usr/bin/env python

import argparse

import globus_sdk
from globus_sdk.experimental.globus_app import UserApp

# Tutorial Client ID - <replace this with your own client>
NATIVE_CLIENT_ID = "61338d24-54d5-408f-a10d-66c06b59f6d2"
USER_APP = UserApp("manage-timers-example", client_id=NATIVE_CLIENT_ID)


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("TIMER_ID")
    args = parser.parse_args()

    client = globus_sdk.TimersClient(app=USER_APP)

    client.delete_job(args.TIMER_ID)
    print("Finished deleting timer.")


if __name__ == "__main__":
    main()



"""
This module provides internal helpers for remarshalling data from one shape
into another.

This may be as simple as converting strings from one form to another, or as
sophisticated as building a specific internal object in a configurable way.
"""

from __future__ import annotations

import collections.abc
import sys
import typing as t
import uuid

from globus_sdk._missing import MISSING, MissingType

if sys.version_info >= (3, 10):
    from typing import TypeAlias
else:
    from typing_extensions import TypeAlias

T = t.TypeVar("T")
R = t.TypeVar("R")

# Omittable[T] means "T may be missing"
# NullableOmittable[T] means "T may be missing or null"
#
# in type systems this kind of construction is sometimes called "Optional" or "Maybe"
# but Python uses "Optional" to mean "T | None" and in the SDK, "None" means "null"
Omittable: TypeAlias[T] = t.Union[T, MissingType]
NullableOmittable: TypeAlias[T] = t.Union[Omittable[T], None]


def stringify(value: NullableOmittable[object]) -> NullableOmittable[str]:
    """
    Convert a value to a string, with handling for None and Missing.

    :param value: The stringifiable object
    """
    if value is None:
        return None
    if value is MISSING:
        return MISSING
    return str(value)


@t.overload
def listify(value: None) -> None: ...
@t.overload
def listify(value: MissingType) -> MissingType: ...
@t.overload
def listify(value: t.Iterable[T]) -> list[T]: ...


def listify(value: NullableOmittable[t.Iterable[T]]) -> NullableOmittable[list[T]]:
    """
    Convert any iterable to a list, with handling for None and Missing.

    :param value: The iterable of objects
    """
    if value is None:
        return None
    if value is MISSING:
        return MISSING
    if isinstance(value, list):
        return value
    return list(value)


def strseq_iter(
    value: t.Iterable[str | uuid.UUID] | str | uuid.UUID,
) -> t.Iterator[str]:
    """
    Iterate over one or more string/string-convertible values.

    :param value: The stringifiable object or objects to iterate over

    This function handles strings, which are themselves iterable,
    by producing the string itself, not it's characters:

    >>> list("foo")
    ['f', 'o', 'o']
    >>> list(strseq_iter("foo"))
    ['foo']

    It also accepts and converts UUIDs and iterables thereof:

    >>> list(strseq_iter(UUID(int=0)))
    ['00000000-0000-0000-0000-000000000000']
    >>> list(strseq_iter([UUID(int=0), UUID(int=1)]))
    ['00000000-0000-0000-0000-000000000000', '00000000-0000-0000-0000-000000000001']

    This helps handle cases where a string is passed to a function expecting an iterable
    of strings, as well as cases where an iterable of UUID objects is accepted for a
    list of IDs, or something similar.
    """
    if isinstance(value, str):
        yield value
    elif isinstance(value, uuid.UUID):
        yield str(value)
    else:
        for x in value:
            yield str(x)


@t.overload
def strseq_listify(value: None) -> None: ...
@t.overload
def strseq_listify(value: MissingType) -> MissingType: ...


@t.overload
def strseq_listify(
    value: t.Iterable[str | uuid.UUID] | str | uuid.UUID,
) -> list[str]: ...


def strseq_listify(
    value: NullableOmittable[t.Iterable[str | uuid.UUID] | str | uuid.UUID],
) -> NullableOmittable[list[str]]:
    """
    A wrapper over strseq_iter which produces list outputs.
    This method takes responsibility for checking for MISSING and None values.

    Unlike strseq_iter, this may be the "last mile" remarshalling step before
    data is actually passed to the network layer. Therefore, it makes sense for this
    helper to handle (MISSING | None).

    :param value: The stringifiable object or iterable of objects
    """
    if value is None:
        return None
    if value is MISSING:
        return MISSING
    return list(strseq_iter(value))


@t.overload
def list_map(value: None, mapped_function: t.Callable[[T], R]) -> None: ...


@t.overload
def list_map(
    value: MissingType, mapped_function: t.Callable[[T], R]
) -> MissingType: ...


@t.overload
def list_map(value: t.Iterable[T], mapped_function: t.Callable[[T], R]) -> list[R]: ...


def list_map(
    value: NullableOmittable[t.Iterable[T]], mapped_function: t.Callable[[T], R]
) -> NullableOmittable[list[R]]:
    """
    Like list(map()) but handles None|MISSING.

    :param value: The iterable of objects over which to map
    :param mapped_function: The function to map
    """
    if value is None:
        return None
    if value is MISSING:
        return MISSING
    return [mapped_function(element) for element in value]


@t.overload
def commajoin(value: MissingType) -> MissingType: ...
@t.overload
def commajoin(value: None) -> None: ...
@t.overload
def commajoin(value: str | uuid.UUID | t.Iterable[str | uuid.UUID]) -> str: ...


def commajoin(
    value: NullableOmittable[str | uuid.UUID | t.Iterable[str | uuid.UUID]],
) -> NullableOmittable[str]:
    if value is None:
        return None
    if value is MISSING:
        return MISSING
    # note that this explicit handling of Iterable allows for objects to be
    # passed to this function and be stringified by the `str()` call
    if isinstance(value, collections.abc.Iterable):
        return ",".join(strseq_iter(value))
    return str(value)



import collections.abc
import uuid

import pytest

from globus_sdk import MISSING
from globus_sdk._internal.remarshal import (
    commajoin,
    list_map,
    listify,
    strseq_iter,
    strseq_listify,
)


@pytest.mark.parametrize(
    "value, expected_result",
    (
        ("foo", ["foo"]),
        ((1, 2, 3), ["1", "2", "3"]),
        (uuid.UUID(int=10), [f"{uuid.UUID(int=10)}"]),
        (["foo", uuid.UUID(int=5)], ["foo", f"{uuid.UUID(int=5)}"]),
    ),
)
def test_strseq_iter(value, expected_result):
    iter_ = strseq_iter(value)
    assert not isinstance(iter_, list)
    assert isinstance(iter_, collections.abc.Iterator)
    assert list(iter_) == expected_result
    assert list(iter_) == []


@pytest.mark.parametrize(
    "value, expected_result",
    (
        ("foo", ["foo"]),
        ((1, 2, 3), ["1", "2", "3"]),
        (uuid.UUID(int=10), [f"{uuid.UUID(int=10)}"]),
        (["foo", uuid.UUID(int=5)], ["foo", f"{uuid.UUID(int=5)}"]),
        (MISSING, MISSING),
        (None, None),
    ),
)
def test_strseq_listify(value, expected_result):
    list_ = strseq_listify(value)
    assert isinstance(list_, list) or list_ in (MISSING, None)
    assert list_ == expected_result


@pytest.mark.parametrize(
    "value, expected_result",
    (
        ("foo", "foo"),
        (uuid.UUID(int=10), f"{uuid.UUID(int=10)}"),
        ((1, 2, 3), "1,2,3"),
        (range(5), "0,1,2,3,4"),
        (["foo", uuid.UUID(int=5)], f"foo,{uuid.UUID(int=5)}"),
        (MISSING, MISSING),
        (None, None),
    ),
)
def test_commajoin(value, expected_result):
    joined = commajoin(value)
    assert joined == expected_result


@pytest.mark.parametrize(
    "value, expected_result",
    (
        ("foo", ["f", "o", "o"]),
        ((1, 2, 3), [1, 2, 3]),
        (range(5), [0, 1, 2, 3, 4]),
        (["foo", uuid.UUID(int=5)], ["foo", uuid.UUID(int=5)]),
        (MISSING, MISSING),
        (None, None),
    ),
)
def test_listify(value, expected_result):
    converted = listify(value)
    assert converted == expected_result


@pytest.mark.parametrize(
    "value, expected_result",
    (
        ("foo", ["ff", "oo", "oo"]),
        ((1, 2, 3), [2, 4, 6]),
        (range(5), [0, 2, 4, 6, 8]),
        (MISSING, MISSING),
        (None, None),
    ),
)
def test_list_map(value, expected_result):
    converted = list_map(value, lambda x: x * 2)
    assert converted == expected_result



from __future__ import annotations

import typing as t
import uuid

from globus_sdk import client, paging, response
from globus_sdk._internal.classprop import classproperty
from globus_sdk._internal.remarshal import commajoin
from globus_sdk._internal.utils import slash_join
from globus_sdk._missing import MISSING, MissingType
from globus_sdk.authorizers import GlobusAuthorizer
from globus_sdk.globus_app import GlobusApp
from globus_sdk.scopes import GCSCollectionScopes, GCSEndpointScopes, Scope
from globus_sdk.transport import RequestsTransport, RetryConfig

from .data import (
    CollectionDocument,
    EndpointDocument,
    GCSRoleDocument,
    StorageGatewayDocument,
    UserCredentialDocument,
)
from .errors import GCSAPIError
from .response import IterableGCSResponse, UnpackingGCSResponse

C = t.TypeVar("C", bound=t.Callable[..., t.Any])


class GCSClient(client.BaseClient):
    """
    A GCSClient provides communication with the GCS Manager API of a Globus Connect
    Server instance.
    For full reference, see the `documentation for the GCS Manager API
    <https://docs.globus.org/globus-connect-server/v5/api/>`_.

    Unlike other client types, this must be provided with an address for the GCS
    Manager. All other arguments are the same as those for
    :class:`~globus_sdk.BaseClient`.

    .. sdk-sphinx-copy-params:: BaseClient

        :param gcs_address: The FQDN (DNS name) or HTTPS URL for the GCS Manager API.

    .. automethodlist:: globus_sdk.GCSClient
    """

    # TODO: under SDK v4.0, service_name should not be set
    service_name = "globus_connect_server"
    error_class = GCSAPIError

    def __init__(
        self,
        gcs_address: str,
        *,
        app: GlobusApp | None = None,
        app_scopes: list[Scope] | None = None,
        environment: str | None = None,
        authorizer: GlobusAuthorizer | None = None,
        app_name: str | None = None,
        transport: RequestsTransport | None = None,
        retry_config: RetryConfig | None = None,
    ) -> None:
        # check if the provided address was a DNS name or an HTTPS URL
        if not gcs_address.startswith("https://"):
            # if it's a DNS name format it accordingly
            gcs_address = f"https://{gcs_address}/api/"
        # if it was an HTTPS URL, check that it ends with /api/
        elif not gcs_address.endswith(("/api/", "/api")):
            # if it doesn't, add it
            gcs_address = slash_join(gcs_address, "/api/")

        self._endpoint_client_id: str | None = None

        super().__init__(
            base_url=gcs_address,
            environment=environment,
            app=app,
            app_scopes=app_scopes,
            authorizer=authorizer,
            app_name=app_name,
            transport=transport,
            retry_config=retry_config,
        )

    @staticmethod
    def get_gcs_endpoint_scopes(
        endpoint_id: uuid.UUID | str,
    ) -> GCSEndpointScopes:
        """Given a GCS Endpoint ID, this helper constructs an object containing the
        scopes for that Endpoint.

        :param endpoint_id: The ID of the Endpoint

        See documentation for :class:`globus_sdk.scopes.GCSEndpointScopes` for
        more information.
        """
        return GCSEndpointScopes(str(endpoint_id))

    @staticmethod
    def get_gcs_collection_scopes(
        collection_id: uuid.UUID | str,
    ) -> GCSCollectionScopes:
        """Given a GCS Collection ID, this helper constructs an object containing the
        scopes for that Collection.

        :param collection_id: The ID of the Collection

        See documentation for :class:`globus_sdk.scopes.GCSCollectionScopes` for
        more information.
        """
        return GCSCollectionScopes(str(collection_id))

    @property
    def default_scope_requirements(self) -> list[Scope]:
        return [
            GCSClient.get_gcs_endpoint_scopes(
                self.endpoint_client_id
            ).manage_collections
        ]

    @classproperty
    def resource_server(  # pylint: disable=missing-param-doc
        self_or_cls: client.BaseClient | type[client.BaseClient],
    ) -> str | None:
        """
        The resource server for a GCS endpoint is the ID of its GCS Manager Client.

        This will return None if called as a classmethod as an instantiated
        ``GCSClient`` is required to look up the client ID from the endpoint.
        """
        if not isinstance(self_or_cls, GCSClient):
            return None

        return self_or_cls.endpoint_client_id

    @property
    def endpoint_client_id(self) -> str:
        """
        The UUID of the GCS Manager client of the endpoint this client is configured
        for. This will be equal to the ``endpoint_id`` in most cases, but when they
        differ the ``client_id`` is the canonical value for the endpoint's resource
        server and scopes.
        """
        if self._endpoint_client_id:
            return self._endpoint_client_id

        else:
            data = self.get_gcs_info()
            try:
                endpoint_id = str(data["client_id"])
            except KeyError:
                print(data)
            self._endpoint_client_id = endpoint_id
            return endpoint_id

    #
    # endpoint methods
    #

    def get_gcs_info(
        self, query_params: dict[str, t.Any] | None = None
    ) -> UnpackingGCSResponse:
        """
        Get information about the GCS Manager service this client is configured for.
        This call is made unauthenticated.

        :param query_params: Additional passthrough query parameters

        .. tab-set::

            .. tab-item:: API Info

                ``GET /info``

                .. extdoclink:: Get Endpoint
                    :ref: openapi/#getInfo
                    :service: gcs
        """
        return UnpackingGCSResponse(
            self.get("/info", query_params=query_params, automatic_authorization=False),
            "info",
        )

    def get_endpoint(
        self, query_params: dict[str, t.Any] | None = None
    ) -> UnpackingGCSResponse:
        """
        Get the details of the Endpoint that this client is configured to talk to.

        :param query_params: Additional passthrough query parameters

        .. tab-set::

            .. tab-item:: API Info

                ``GET /endpoint``

                .. extdoclink:: Get Endpoint
                    :ref: openapi_Endpoint/#getEndpoint
                    :service: gcs
        """
        return UnpackingGCSResponse(
            self.get("/endpoint", query_params=query_params),
            "endpoint",
        )

    def update_endpoint(
        self,
        endpoint_data: dict[str, t.Any] | EndpointDocument,
        *,
        include: (
            t.Iterable[t.Literal["endpoint"]] | t.Literal["endpoint"] | MissingType
        ) = MISSING,
        query_params: dict[str, t.Any] | None = None,
    ) -> UnpackingGCSResponse:
        """
        Update a GCSv5 Endpoint

        :param endpoint_data: The endpoint document for the modified endpoint
        :param include: Optional list of document types to include in the response
           (currently only supports the value ``["endpoint"]``)
        :param query_params: Additional passthrough query parameters

        .. tab-set::

            .. tab-item:: API Info

                ``PATCH /endpoint``

                .. extdoclink:: Update Endpoint
                    :ref: openapi_Endpoint/#patchEndpoint
                    :service: gcs
        """
        query_params = {"include": commajoin(include), **(query_params or {})}
        return UnpackingGCSResponse(
            self.patch(
                "/endpoint",
                data=endpoint_data,
                query_params=query_params,
            ),
            "endpoint",
        )

    #
    # collection methods
    #

    @paging.has_paginator(
        paging.MarkerPaginator,
        items_key="data",
    )
    def get_collection_list(
        self,
        *,
        mapped_collection_id: uuid.UUID | str | MissingType = MISSING,
        filter: (  # pylint: disable=redefined-builtin
            str | t.Iterable[str] | MissingType
        ) = MISSING,
        include: str | t.Iterable[str] | MissingType = MISSING,
        page_size: int | MissingType = MISSING,
        marker: str | MissingType = MISSING,
        query_params: dict[str, t.Any] | None = None,
    ) -> IterableGCSResponse:
        """
        List the Collections on an Endpoint

        :param mapped_collection_id: Filter collections which were created using this
            mapped collection ID.
        :param filter: Filter the returned set to any combination of the following:
            ``mapped_collections``, ``guest_collections``, ``managed_by_me``,
            ``created_by_me``.
        :param include: Names of additional documents to include in the response
        :param page_size: Number of results to return per page
        :param marker: Pagination marker supplied by previous API calls in the event
            a request returns more values than the page size
        :param query_params: Additional passthrough query parameters

        .. tab-set::

            .. tab-item:: API Info

                ``GET /collections``

                .. extdoclink:: List Collections
                    :ref: openapi_Collections/#ListCollections
                    :service: gcs
        """
        query_params = {
            "include": commajoin(include),
            "page_size": page_size,
            "marker": marker,
            "mapped_collection_id": mapped_collection_id,
            "filter": commajoin(filter),
            **(query_params or {}),
        }
        return IterableGCSResponse(self.get("collections", query_params=query_params))

    def get_collection(
        self,
        collection_id: uuid.UUID | str,
        *,
        query_params: dict[str, t.Any] | None = None,
    ) -> UnpackingGCSResponse:
        """
        Lookup a Collection on an Endpoint

        :param collection_id: The ID of the collection to lookup
        :param query_params: Additional passthrough query parameters

        .. tab-set::

            .. tab-item:: API Info

                ``GET /collections/{collection_id}``

                .. extdoclink:: Get Collection
                    :ref: openapi_Collections/#getCollection
                    :service: gcs
        """
        return UnpackingGCSResponse(
            self.get(f"/collections/{collection_id}", query_params=query_params),
            "collection",
        )

    def create_collection(
        self,
        collection_data: dict[str, t.Any] | CollectionDocument,
    ) -> UnpackingGCSResponse:
        """
        Create a collection. This is used to create either a mapped or a guest
        collection. When created, a ``collection:administrator`` role for that
        collection will be created using the caller’s identity.

        In order to create a guest collection, the caller must have an identity that
        matches the Storage Gateway policies.

        In order to create a mapped collection, the caller must have an
        ``endpoint:administrator`` or ``endpoint:owner`` role.

        :param collection_data: The collection document for the new collection

        .. tab-set::

            .. tab-item:: API Info

                ``POST /collections``

                .. extdoclink:: Create Collection
                    :ref: openapi_Collections/#createCollection
                    :service: gcs
        """
        return UnpackingGCSResponse(
            self.post("/collections", data=collection_data), "collection"
        )

    def update_collection(
        self,
        collection_id: uuid.UUID | str,
        collection_data: dict[str, t.Any] | CollectionDocument,
        *,
        query_params: dict[str, t.Any] | None = None,
    ) -> UnpackingGCSResponse:
        """
        Update a Collection

        :param collection_id: The ID of the collection to update
        :param collection_data: The collection document for the modified collection
        :param query_params: Additional passthrough query parameters

        .. tab-set::

            .. tab-item:: API Info

                ``PATCH /collections/{collection_id}``

                .. extdoclink:: Update Collection
                    :ref: openapi_Collections/#patchCollection
                    :service: gcs
        """
        return UnpackingGCSResponse(
            self.patch(
                f"/collections/{collection_id}",
                data=collection_data,
                query_params=query_params,
            ),
            "collection",
        )

    def delete_collection(
        self,
        collection_id: uuid.UUID | str,
        *,
        query_params: dict[str, t.Any] | None = None,
    ) -> response.GlobusHTTPResponse:
        """
        Delete a Collection

        :param collection_id: The ID of the collection to delete
        :param query_params: Additional passthrough query parameters

        .. tab-set::

            .. tab-item:: API Info

                ``DELETE /collections/{collection_id}``

                .. extdoclink:: Delete Collection
                    :ref: openapi_Collections/#deleteCollection
                    :service: gcs
        """
        return self.delete(f"/collections/{collection_id}", query_params=query_params)

    #
    # storage gateway methods
    #

    @paging.has_paginator(
        paging.MarkerPaginator,
        items_key="data",
    )
    def get_storage_gateway_list(
        self,
        *,
        include: str | t.Iterable[str] | MissingType = MISSING,
        page_size: int | MissingType = MISSING,
        marker: str | MissingType = MISSING,
        query_params: dict[str, t.Any] | None = None,
    ) -> IterableGCSResponse:
        """
        List Storage Gateways

        :param include: Optional document types to include in the response. If
            'private_policies' is included, then include private storage gateway
            policies in the attached storage_gateways document. This requires an
            ``administrator`` role on the Endpoint.
        :param page_size: Number of results to return per page
        :param marker: Pagination marker supplied by previous API calls in the event
            a request returns more values than the page size
        :param query_params: Additional passthrough query parameters

        .. tab-set::

            .. tab-item:: Paginated Usage

                .. paginatedusage:: get_storage_gateway_list

            .. tab-item:: API Info

                ``GET /storage_gateways``

                .. extdoclink:: Delete Collection
                    :ref: openapi_Storage_Gateways/#getStorageGateways
                    :service: gcs
        """
        query_params = {
            "include": commajoin(include),
            "page_size": page_size,
            "marker": marker,
            **(query_params or {}),
        }
        return IterableGCSResponse(
            self.get("/storage_gateways", query_params=query_params)
        )

    def create_storage_gateway(
        self,
        data: dict[str, t.Any] | StorageGatewayDocument,
        *,
        query_params: dict[str, t.Any] | None = None,
    ) -> UnpackingGCSResponse:
        """
        Create a Storage Gateway

        :param data: Data in the format of a Storage Gateway document, it is recommended
            to use the ``StorageGatewayDocument`` class to construct this data.
        :param query_params: Additional passthrough query parameters

        .. tab-set::

            .. tab-item:: API Info

                ``POST /storage_gateways``

                .. extdoclink:: Create Storage Gateway
                    :ref: openapi_Storage_Gateways/#postStorageGateway
                    :service: gcs
        """
        return UnpackingGCSResponse(
            self.post("/storage_gateways", data=data, query_params=query_params),
            "storage_gateway",
        )

    def get_storage_gateway(
        self,
        storage_gateway_id: uuid.UUID | str,
        *,
        include: str | t.Iterable[str] | MissingType = MISSING,
        query_params: dict[str, t.Any] | None = None,
    ) -> UnpackingGCSResponse:
        """
        Lookup a Storage Gateway by ID

        :param storage_gateway_id: UUID for the Storage Gateway to be gotten
        :param include: Optional document types to include in the response. If
            'private_policies' is included, then include private storage gateway
            policies in the attached storage_gateways document. This requires an
            ``administrator`` role on the Endpoint.
        :param query_params: Additional passthrough query parameters

        .. tab-set::

            .. tab-item:: API Info

                ``GET /storage_gateways/<storage_gateway_id>``

                .. extdoclink:: Get a Storage Gateway
                    :ref: openapi_Storage_Gateways/#getStorageGateway
                    :service: gcs
        """
        query_params = {"include": commajoin(include), **(query_params or {})}
        return UnpackingGCSResponse(
            self.get(
                f"/storage_gateways/{storage_gateway_id}",
                query_params=query_params,
            ),
            "storage_gateway",
        )

    def update_storage_gateway(
        self,
        storage_gateway_id: uuid.UUID | str,
        data: dict[str, t.Any] | StorageGatewayDocument,
        *,
        query_params: dict[str, t.Any] | None = None,
    ) -> response.GlobusHTTPResponse:
        """
        Update a Storage Gateway

        :param storage_gateway_id: UUID for the Storage Gateway to be updated
        :param data: Data in the format of a Storage Gateway document, it is recommended
            to use the ``StorageGatewayDocument`` class to construct this data.
        :param query_params: Additional passthrough query parameters

        .. tab-set::

            .. tab-item:: API Info

                ``PATCH /storage_gateways/<storage_gateway_id>``

                .. extdoclink:: Update a Storage Gateway
                    :ref: openapi_Storage_Gateways/#patchStorageGateway
                    :service: gcs
        """
        return self.patch(
            f"/storage_gateways/{storage_gateway_id}",
            data=data,
            query_params=query_params,
        )

    def delete_storage_gateway(
        self,
        storage_gateway_id: str | uuid.UUID,
        *,
        query_params: dict[str, t.Any] | None = None,
    ) -> response.GlobusHTTPResponse:
        """
        Delete a Storage Gateway

        :param storage_gateway_id: UUID for the Storage Gateway to be deleted
        :param query_params: Additional passthrough query parameters

        .. tab-set::

            .. tab-item:: API Info

                ``DELETE /storage_gateways/<storage_gateway_id>``

                .. extdoclink:: Delete a Storage Gateway
                    :ref: openapi_Storage_Gateways/#deleteStorageGateway
                    :service: gcs
        """
        return self.delete(
            f"/storage_gateways/{storage_gateway_id}", query_params=query_params
        )

    #
    # role methods
    #

    @paging.has_paginator(
        paging.MarkerPaginator,
        items_key="data",
    )
    def get_role_list(
        self,
        collection_id: uuid.UUID | str | MissingType = MISSING,
        include: str | MissingType = MISSING,
        page_size: int | MissingType = MISSING,
        marker: str | MissingType = MISSING,
        query_params: dict[str, t.Any] | None = None,
    ) -> IterableGCSResponse:
        """
        List Roles

        :param collection_id: UUID of a Collection. If given then only roles
            related to that Collection are returned, otherwise only Endpoint
            roles are returned.
        :param include: Pass "all_roles" to request all roles
            relevant to the resource instead of only those the caller has on
            the resource
        :param page_size: Number of results to return per page
        :param marker: Pagination marker supplied by previous API calls in the event
            a request returns more values than the page size
        :param query_params: Additional passthrough query parameters

        .. tab-set::

            .. tab-item:: API Info

                ``GET /roles``

                .. extdoclink:: Delete a Storage Gateway
                    :ref: openapi_Roles/#listRoles
                    :service: gcs
        """
        query_params = {
            "include": include,
            "page_size": page_size,
            "marker": marker,
            "collection_id": collection_id,
            **(query_params or {}),
        }
        path = "/roles"
        return IterableGCSResponse(self.get(path, query_params=query_params))

    def create_role(
        self,
        data: dict[str, t.Any] | GCSRoleDocument,
        query_params: dict[str, t.Any] | None = None,
    ) -> UnpackingGCSResponse:
        """
        Create a Role

        :param data: Data in the format of a Role document, it is recommended
            to use the `GCSRoleDocument` class to construct this data.
        :param query_params: Additional passthrough query parameters

        .. tab-set::

            .. tab-item:: API Info

                ``POST /roles``

                .. extdoclink:: Create Role
                    :ref: openapi_Roles/#postRole
                    :service: gcs
        """
        path = "/roles"
        return UnpackingGCSResponse(
            self.post(path, data=data, query_params=query_params),
            "role",
        )

    def get_role(
        self,
        role_id: uuid.UUID | str,
        query_params: dict[str, t.Any] | None = None,
    ) -> UnpackingGCSResponse:
        """
        Get a Role by ID

        :param role_id: UUID for the Role to be gotten
        :param query_params: Additional passthrough query parameters

        .. tab-set::

            .. tab-item:: API Info

                ``GET /roles/{role_id}``

                .. extdoclink:: Get Role
                    :ref: openapi_Roles/#getRole
                    :service: gcs
        """
        path = f"/roles/{role_id}"
        return UnpackingGCSResponse(self.get(path, query_params=query_params), "role")

    def delete_role(
        self,
        role_id: uuid.UUID | str,
        query_params: dict[str, t.Any] | None = None,
    ) -> response.GlobusHTTPResponse:
        """
        Delete a Role

        :param role_id: UUID for the Role to be deleted
        :param query_params: Additional passthrough query parameters

        .. tab-set::

            .. tab-item:: API Info

                ``DELETE /roles/{role_id}``

                .. extdoclink:: Delete Role
                    :ref: openapi_Roles/#deleteRole
                    :service: gcs
        """
        path = f"/roles/{role_id}"
        return self.delete(path, query_params=query_params)

    @paging.has_paginator(
        paging.MarkerPaginator,
        items_key="data",
    )
    def get_user_credential_list(
        self,
        storage_gateway: uuid.UUID | str | MissingType = MISSING,
        page_size: int | MissingType = MISSING,
        marker: str | MissingType = MISSING,
        query_params: dict[str, t.Any] | None = None,
    ) -> IterableGCSResponse:
        """
        List User Credentials

        :param storage_gateway: UUID of a storage gateway to limit results to
        :param query_params: Additional passthrough query parameters
        :param page_size: Number of results to return per page
        :param marker: Pagination marker supplied by previous API calls in the event
            a request returns more values than the page size

        .. tab-set::

            .. tab-item:: API Info

                ``GET /user_credentials``

                .. extdoclink:: Get User Credential List
                    :ref: openapi_User_Credentials/#getUserCredentials
                    :service: gcs
        """
        query_params = {
            "storage_gateway": storage_gateway,
            "page_size": page_size,
            "marker": marker,
            **(query_params or {}),
        }
        path = "/user_credentials"
        return IterableGCSResponse(self.get(path, query_params=query_params))

    def create_user_credential(
        self,
        data: dict[str, t.Any] | UserCredentialDocument,
        query_params: dict[str, t.Any] | None = None,
    ) -> UnpackingGCSResponse:
        """
        Create a User Credential

        :param data: Data in the format of a UserCredential document, it is
            recommended to use the `UserCredential` class to construct this
        :param query_params: Additional passthrough query parameters

        .. tab-set::

            .. tab-item:: API Info

                ``POST /user_credentials``

                .. extdoclink:: Create User Credential
                    :ref: openapi_User_Credentials/#postUserCredential
                    :service: gcs
        """
        path = "/user_credentials"
        return UnpackingGCSResponse(
            self.post(path, data=data, query_params=query_params),
            "user_credential",
        )

    def get_user_credential(
        self,
        user_credential_id: uuid.UUID | str,
        query_params: dict[str, t.Any] | None = None,
    ) -> UnpackingGCSResponse:
        """
        Get a User Credential by ID

        :param user_credential_id: UUID for the UserCredential to be gotten
        :param query_params: Additional passthrough query parameters

        .. tab-set::

            .. tab-item:: API Info

                ``GET /user_credentials/{user_credential_id}``

                .. extdoclink:: Get a User Credential
                    :ref: openapi_User_Credentials/#getUserCredential
                    :service: gcs
        """
        path = f"/user_credentials/{user_credential_id}"
        return UnpackingGCSResponse(
            self.get(path, query_params=query_params), "user_credential"
        )

    def update_user_credential(
        self,
        user_credential_id: uuid.UUID | str,
        data: dict[str, t.Any] | UserCredentialDocument,
        query_params: dict[str, t.Any] | None = None,
    ) -> UnpackingGCSResponse:
        """
        Update a User Credential

        :param user_credential_id: UUID for the UserCredential to be updated
        :param data: Data in the format of a UserCredential document, it is
            recommended to use the `UserCredential` class to construct this
        :param query_params: Additional passthrough query parameters

        .. tab-set::

            .. tab-item:: API Info

                ``PATCH /user_credentials/{user_credential_id}``

                .. extdoclink:: Update a User Credential
                    :ref: openapi_User_Credentials/#patchUserCredential
                    :service: gcs
        """
        path = f"/user_credentials/{user_credential_id}"
        return UnpackingGCSResponse(
            self.patch(path, data=data, query_params=query_params), "user_credential"
        )

    def delete_user_credential(
        self,
        user_credential_id: uuid.UUID | str,
        query_params: dict[str, t.Any] | None = None,
    ) -> response.GlobusHTTPResponse:
        """
        Delete a User Credential

        :param user_credential_id: UUID for the UserCredential to be deleted
        :param query_params: Additional passthrough query parameters

        .. tab-set::

            .. tab-item:: API Info

                ``DELETE /user_credentials/{user_credential_id}``

                .. extdoclink:: Delete User Credential
                    :ref: openapi_User_Credentials/#deleteUserCredential
                    :service: gcs
        """
        path = f"/user_credentials/{user_credential_id}"
        return self.delete(path, query_params=query_params)



import urllib.parse

import pytest

from globus_sdk.testing import get_last_request, load_response


@pytest.mark.parametrize(
    "client_kwargs, qs",
    [
        ({}, {}),
        ({"query_params": {"foo": "bar"}}, {"foo": "bar"}),
        ({"filter": "foo"}, {"filter": "foo"}),
        ({"limit": 10, "offset": 100}, {"limit": "10", "offset": "100"}),
        ({"limit": 10, "query_params": {"limit": 100}}, {"limit": "100"}),
        ({"filter": "foo:bar:baz"}, {"filter": "foo:bar:baz"}),
        ({"filter": {"foo": "bar", "bar": "baz"}}, {"filter": "foo:bar/bar:baz"}),
        ({"filter": {"foo": ["bar", "baz"]}}, {"filter": "foo:bar,baz"}),
    ],
)
def test_task_list(client, client_kwargs, qs):
    load_response(client.task_list)
    client.task_list(**client_kwargs)

    req = get_last_request()
    parsed_qs = urllib.parse.parse_qs(urllib.parse.urlparse(req.url).query)
    # parsed_qs will have each value as a list (because query-params are a multidict)
    # so transform the test data to match before comparison
    assert parsed_qs == {k: [v] for k, v in qs.items()}


@pytest.mark.parametrize(
    "orderby_value, expected_orderby_param",
    [
        ("foo", "foo"),
        (["foo"], "foo"),
        ("foo,bar", "foo,bar"),
        ("foo ASC,bar", "foo ASC,bar"),
        (["foo ASC", "bar"], "foo ASC,bar"),
        (["foo ASC", "bar DESC"], "foo ASC,bar DESC"),
    ],
)
def test_task_list_orderby_parameter(client, orderby_value, expected_orderby_param):
    load_response(client.task_list)
    client.task_list(orderby=orderby_value)

    req = get_last_request()
    parsed_qs = urllib.parse.parse_qs(urllib.parse.urlparse(req.url).query)
    assert "orderby" in parsed_qs
    assert len(parsed_qs["orderby"]) == 1
    orderby_param = parsed_qs["orderby"][0]
    assert orderby_param == expected_orderby_param



# this script should only be called via
#    tox run -e check-min-python-is-tested
#
# no other usages are supported
import pathlib
import sys

import mddj.api
import ruamel.yaml

dj = mddj.api.DJ()
YAML = ruamel.yaml.YAML(typ="safe")
REPO_ROOT = pathlib.Path(__file__).parent.parent

requires_python_version = dj.read.requires_python(lower_bound=True)
print("requires-python:", requires_python_version)

with open(REPO_ROOT / ".github" / "workflows" / "test.yaml") as f:
    workflow = YAML.load(f)
    includes = workflow["jobs"]["test"]["strategy"]["matrix"]["include"]
    for include in includes:
        if include["name"] == "Linux":
            break
    else:
        raise ValueError("Could not find 'Linux' in the test matrix.")

    for environment in include["tox-post-environments"]:
        if environment.endswith("-mindeps"):
            break
    else:
        raise ValueError("Could not find a '-mindeps' tox-post-environment.")

    python_version, _, _ = environment.partition("-")
    print("test-mindeps job python:", python_version)
    if python_version != f"py{requires_python_version}":
        print("ERROR: ensure_min_python_is_tested.py failed!")
        print(
            f"\nPackage data sets 'Requires-Python: >={requires_python_version}', "
            f"but the test-mindeps job is configured to test '{python_version}'.\n",
            file=sys.stderr,
        )
        sys.exit(1)


tox_min_python_version = dj.read.tox.min_python_version()
print("tox min python version:", tox_min_python_version)
if tox_min_python_version != requires_python_version:
    print("ERROR: ensure_min_python_is_tested.py failed!")
    print(
        f"\nPackage data sets 'Requires-Python: >={requires_python_version}', "
        "but tox is configured to test with a minimum of "
        f"'{tox_min_python_version}'.\n",
        file=sys.stderr,
    )
    sys.exit(1)



import globus_sdk

# ok usage
ac = globus_sdk.AuthClient()
nc = globus_sdk.NativeAppAuthClient("foo_client_id")
cc = globus_sdk.ConfidentialAppAuthClient("foo_client_id", "foo_client_secret")

# base class allows authorizer
authorizer = globus_sdk.AccessTokenAuthorizer("dummytoken")
ac = globus_sdk.AuthClient(authorizer=authorizer)

# subclasses forbid authorizers
nc = globus_sdk.NativeAppAuthClient(  # type: ignore[call-arg]
    "foo_client_id", authorizer=authorizer
)
cc = globus_sdk.ConfidentialAppAuthClient(  # type: ignore[call-arg]
    "foo_client_id", "foo_client_secret", authorizer=authorizer
)

# the login clients allow a client_id kwarg, but AuthClient does not
globus_sdk.NativeAppAuthClient(client_id="foo_client_id")
globus_sdk.ConfidentialAppAuthClient(
    client_id="foo_client_id", client_secret="foo_client_secret"
)
globus_sdk.AuthClient(client_id="foo_client_id")  # type: ignore[call-arg]



from globus_sdk.testing.models import RegisteredResponse, ResponseSet

from ._common import ENDPOINT_ID

RESPONSES = ResponseSet(
    metadata={"endpoint_id": ENDPOINT_ID},
    default=RegisteredResponse(
        service="transfer",
        method="GET",
        path=f"/v0.10/operation/endpoint/{ENDPOINT_ID}/stat",
        json={
            "DATA_TYPE": "file",
            "group": "tutorial",
            "last_modified": "2023-12-18 16:52:50+00:00",
            "link_group": None,
            "link_last_modified": None,
            "link_size": None,
            "link_target": None,
            "link_user": None,
            "name": "file1.txt",
            "permissions": "0644",
            "size": 4,
            "type": "file",
            "user": "tutorial",
        },
    ),
    not_found=RegisteredResponse(
        service="transfer",
        method="GET",
        path=f"/v0.10/operation/endpoint/{ENDPOINT_ID}/stat",
        status=404,
        json={
            "code": "NotFound",
            "message": f"Path not found, Error (list)\nEndpoint: Globus Tutorial Collection 1 ({ENDPOINT_ID})\nServer: 100.26.231.26:443\nMessage: No such file or directory\n---\nDetails: Error: '~/foo' not found\\r\\n550-GlobusError: v=1 c=PATH_NOT_FOUND\\r\\n550-GridFTP-Errno: 2\\r\\n550-GridFTP-Reason: System error in stat\\r\\n550-GridFTP-Error-String: No such file or directory\\r\\n550 End.\\r\\n\n",  # noqa 501
            "request_id": "aaabbbccc",
            "resource": f"/v0.10/operation/endpoint/{ENDPOINT_ID}/stat",
        },
    ),
    permission_denied=RegisteredResponse(
        service="transfer",
        method="GET",
        path=f"/v0.10/operation/endpoint/{ENDPOINT_ID}/stat",
        status=403,
        json={
            "code": "EndpointPermissionDenied",
            "message": f"Denied by endpoint, Error (list)\nEndpoint: Globus Tutorial Collection 1 ({ENDPOINT_ID})\nServer: 100.26.231.26:443\nCommand: MLST /foo\nMessage: Fatal FTP Response\n---\nDetails: 500 Command failed : Path not allowed.\\r\\n\n",  # noqa 501
            "request_id": "aaabbbccc",
            "resource": f"/v0.10/operation/endpoint/{ENDPOINT_ID}/stat",
        },
    ),
)



from ._missing import MISSING, MissingType
from .authorizers import (
    AccessTokenAuthorizer,
    BasicAuthorizer,
    ClientCredentialsAuthorizer,
    NullAuthorizer,
    RefreshTokenAuthorizer,
)
from .client import BaseClient
from .exc import (
    ErrorSubdocument,
    GlobusAPIError,
    GlobusConnectionError,
    GlobusConnectionTimeoutError,
    GlobusError,
    GlobusSDKUsageError,
    GlobusTimeoutError,
    NetworkError,
    RemovedInV5Warning,
    ValidationError,
)
from .globus_app import ClientApp, GlobusApp, GlobusAppConfig, UserApp
from .local_endpoint import (
    GlobusConnectPersonalOwnerInfo,
    LocalGlobusConnectPersonal,
    LocalGlobusConnectServer,
)
from .response import ArrayResponse, GlobusHTTPResponse, IterableResponse
from .scopes import Scope, ScopeCycleError, ScopeParseError
from .services.auth import (
    AuthAPIError,
    AuthClient,
    AuthLoginClient,
    ConfidentialAppAuthClient,
    DependentScopeSpec,
    GetConsentsResponse,
    GetIdentitiesResponse,
    IdentityMap,
    IDTokenDecoder,
    NativeAppAuthClient,
    OAuthAuthorizationCodeResponse,
    OAuthClientCredentialsResponse,
    OAuthDependentTokenResponse,
    OAuthRefreshTokenResponse,
    OAuthTokenResponse,
)
from .services.compute import (
    ComputeAPIError,
    ComputeClientV2,
    ComputeClientV3,
)
from .services.flows import (
    FlowsAPIError,
    FlowsClient,
    IterableFlowsResponse,
    RunActivityNotificationPolicy,
    SpecificFlowClient,
)
from .services.gcs import (
    ActiveScaleStoragePolicies,
    AzureBlobStoragePolicies,
    BlackPearlStoragePolicies,
    BoxStoragePolicies,
    CephStoragePolicies,
    CollectionDocument,
    CollectionPolicies,
    ConnectorTable,
    EndpointDocument,
    GCSAPIError,
    GCSClient,
    GCSRoleDocument,
    GlobusConnectServerConnector,
    GoogleCloudStorageCollectionPolicies,
    GoogleCloudStoragePolicies,
    GoogleDriveStoragePolicies,
    GuestCollectionDocument,
    HPSSStoragePolicies,
    IrodsStoragePolicies,
    IterableGCSResponse,
    MappedCollectionDocument,
    OneDriveStoragePolicies,
    POSIXCollectionPolicies,
    POSIXStagingCollectionPolicies,
    POSIXStagingStoragePolicies,
    POSIXStoragePolicies,
    S3StoragePolicies,
    StorageGatewayDocument,
    StorageGatewayPolicies,
    UnpackingGCSResponse,
    UserCredentialDocument,
)
from .services.groups import (
    BatchMembershipActions,
    GroupMemberVisibility,
    GroupPolicies,
    GroupRequiredSignupFields,
    GroupRole,
    GroupsAPIError,
    GroupsClient,
    GroupsManager,
    GroupVisibility,
)
from .services.search import (
    SearchAPIError,
    SearchClient,
    SearchQueryV1,
    SearchScrollQuery,
)
from .services.timers import (
    FlowTimer,
    OnceTimerSchedule,
    RecurringTimerSchedule,
    TimerJob,
    TimersAPIError,
    TimersClient,
    TransferTimer,
)
from .services.transfer import (
    CreateTunnelData,
    DeleteData,
    IterableTransferResponse,
    TransferAPIError,
    TransferClient,
    TransferData,
)

__version__ = "x.y.z"

def _force_eager_imports() -> None: ...

__all__ = (
    "AccessTokenAuthorizer",
    "BasicAuthorizer",
    "ClientCredentialsAuthorizer",
    "NullAuthorizer",
    "RefreshTokenAuthorizer",
    "BaseClient",
    "ErrorSubdocument",
    "GlobusAPIError",
    "GlobusConnectionError",
    "GlobusConnectionTimeoutError",
    "GlobusError",
    "GlobusSDKUsageError",
    "GlobusTimeoutError",
    "NetworkError",
    "RemovedInV5Warning",
    "ValidationError",
    "ClientApp",
    "GlobusApp",
    "GlobusAppConfig",
    "UserApp",
    "GlobusConnectPersonalOwnerInfo",
    "LocalGlobusConnectPersonal",
    "LocalGlobusConnectServer",
    "ArrayResponse",
    "GlobusHTTPResponse",
    "IterableResponse",
    "Scope",
    "ScopeCycleError",
    "ScopeParseError",
    "AuthAPIError",
    "AuthClient",
    "AuthLoginClient",
    "ConfidentialAppAuthClient",
    "DependentScopeSpec",
    "GetConsentsResponse",
    "GetIdentitiesResponse",
    "IdentityMap",
    "NativeAppAuthClient",
    "OAuthAuthorizationCodeResponse",
    "OAuthClientCredentialsResponse",
    "OAuthDependentTokenResponse",
    "OAuthRefreshTokenResponse",
    "OAuthTokenResponse",
    "IDTokenDecoder",
    "ComputeAPIError",
    "ComputeClientV2",
    "ComputeClientV3",
    "FlowsAPIError",
    "FlowsClient",
    "IterableFlowsResponse",
    "RunActivityNotificationPolicy",
    "SpecificFlowClient",
    "ActiveScaleStoragePolicies",
    "AzureBlobStoragePolicies",
    "BlackPearlStoragePolicies",
    "BoxStoragePolicies",
    "CephStoragePolicies",
    "CollectionDocument",
    "CollectionPolicies",
    "ConnectorTable",
    "EndpointDocument",
    "GCSAPIError",
    "GCSClient",
    "GCSRoleDocument",
    "GlobusConnectServerConnector",
    "GoogleCloudStorageCollectionPolicies",
    "GoogleCloudStoragePolicies",
    "GoogleDriveStoragePolicies",
    "GuestCollectionDocument",
    "HPSSStoragePolicies",
    "IrodsStoragePolicies",
    "IterableGCSResponse",
    "MappedCollectionDocument",
    "OneDriveStoragePolicies",
    "POSIXCollectionPolicies",
    "POSIXStagingCollectionPolicies",
    "POSIXStagingStoragePolicies",
    "POSIXStoragePolicies",
    "S3StoragePolicies",
    "StorageGatewayDocument",
    "StorageGatewayPolicies",
    "UnpackingGCSResponse",
    "UserCredentialDocument",
    "BatchMembershipActions",
    "GroupMemberVisibility",
    "GroupPolicies",
    "GroupRequiredSignupFields",
    "GroupRole",
    "GroupsAPIError",
    "GroupsClient",
    "GroupsManager",
    "GroupVisibility",
    "SearchAPIError",
    "SearchClient",
    "SearchQueryV1",
    "SearchScrollQuery",
    "OnceTimerSchedule",
    "RecurringTimerSchedule",
    "TimerJob",
    "TimersAPIError",
    "TimersClient",
    "FlowTimer",
    "TransferTimer",
    "DeleteData",
    "IterableTransferResponse",
    "TransferAPIError",
    "TransferClient",
    "TransferData",
    "MISSING",
    "MissingType",
    "__version__",
    "_force_eager_imports",
)



import globus_sdk
from globus_sdk.testing import load_response


def test_get_version(compute_client_v2: globus_sdk.ComputeClientV2):
    meta = load_response(compute_client_v2.get_version).metadata
    res = compute_client_v2.get_version()
    assert res.http_status == 200
    assert res.data == meta["api_version"]


def test_get_version_all(compute_client_v2: globus_sdk.ComputeClientV2):
    meta = load_response(compute_client_v2.get_version, case="all").metadata
    res = compute_client_v2.get_version(service="all")
    assert res.http_status == 200
    assert res.data["api"] == meta["api_version"]



import uuid

from globus_sdk import TransferData

# simple usage, ok
TransferData("srcep", "destep")

# can set sync level
TransferData("srcep", "destep", sync_level=1)
TransferData("srcep", "destep", sync_level="exists")
# unknown int values are allowed
TransferData("srcep", "destep", sync_level=100)
# unknown str values are rejected (Literal)
TransferData("srcep", "destep", sync_level="sizes")  # type: ignore[arg-type]

# TransferData.add_filter_rule
tdata = TransferData(uuid.UUID(), uuid.UUID())
tdata.add_filter_rule("*.tgz")
tdata.add_filter_rule("*.tgz", method="exclude")
tdata.add_filter_rule("*.tgz", type="file")
# bad values rejected (Literal)
tdata.add_filter_rule("*.tgz", type="files")  # type: ignore[arg-type]
tdata.add_filter_rule("*.tgz", method="occlude")  # type: ignore[arg-type]



import uuid

from globus_sdk.testing.models import RegisteredResponse, ResponseSet

BOOKMARK_ID = str(uuid.uuid4())

_collection_id = str(uuid.uuid4())
_name = "public datasets"
_path = "/data_repository/public"


RESPONSES = ResponseSet(
    default=RegisteredResponse(
        service="transfer",
        method="POST",
        path="/v2/bookmarks",
        status=201,
        json={
            "meta": {"request_id": "nDZoeEMPI"},
            "data": {
                "type": "Bookmark",
                "id": BOOKMARK_ID,
                "attributes": {
                    "name": _name,
                    "path": _path,
                    "pinned": False,
                },
                "relationships": {
                    "collection": {
                        "data": {
                            "type": "Collection",
                            "id": _collection_id,
                        }
                    }
                },
            },
        },
        metadata={
            "bookmark_id": BOOKMARK_ID,
            "collection": _collection_id,
            "name": _name,
            "path": _path,
        },
    )
)
